import mock
import re
from django.test import TestCase
import responses
from django.contrib.auth.models import User
from django.test.client import RequestFactory
from dbmi_client import authz
from dbmi_client.settings import dbmi_settings


class TestAuthz(TestCase):

    # User JWT
    fake_jwt = "somefakejwtalsdijlaiwjdlijawlidjasdhgashgd"

    @classmethod
    def setUpClass(cls):

        # Patch JWT checker
        cls.jwt_patcher = mock.patch("dbmi_client.authn.validate_request")
        cls.mock_jwt = cls.jwt_patcher.start()
        cls.mock_jwt.return_value = True

        # Set a user and request factory
        cls.rf = RequestFactory()
        cls.user = User.objects.create_user(username="testuser", email="testuser@dbmiauth.local", password="top_secret")

        # Set the authorization server URL
        cls.authz_url_pattern = re.compile(dbmi_settings.AUTHZ_URL + r".*")

    @classmethod
    def tearDownClass(cls):

        # Disable patcher
        cls.jwt_patcher.stop()

    def build_request(self, path, method="get"):

        # Build the request
        request = getattr(self.rf, method)(path, HTTP_AUTHORIZATION=TestAuthz.fake_jwt)

        # Add the user
        request.user = self.user
        request.COOKIES[dbmi_settings.JWT_COOKIE_NAME] = "JWT {}".format(TestAuthz.fake_jwt)

        return request

    def setUp(self):

        # Setup each test
        pass

    @responses.activate
    @mock.patch("dbmi_client.authz.has_a_permission")
    def test_authz_mock_(self, mock_has_a_permission):

        # Set the response
        mock_has_a_permission.return_value = True

        # Build the request
        request = self.build_request("/some/page/")

        # Run it
        has_authz = authz.has_a_permission(request, self.user.email, "item", ["admin", "read"], check_parents=True)

        # Check it
        self.assertTrue(has_authz)

    @responses.activate
    def test_authz_no_response(self):

        # Build the request
        request = self.build_request("/some/page/")

        # Build the response handler
        responses.add(responses.GET, self.authz_url_pattern, json={"error": "not found"}, status=404)

        # Build the call
        has_authz = authz.has_permission(request, self.user.email, "item", "admin", check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertFalse(has_authz)

    @responses.activate
    def test_authz_no_perm(self):

        # Build the request
        request = self.build_request("/some/page/")

        # Build the response handler
        responses.add(
            responses.GET,
            self.authz_url_pattern,
            json={"results": [{"item": "someitem", "permission": "admin", "email": self.user.email}]},
            status=200,
        )

        # Build the call
        has_authz = authz.has_permission(request, self.user.email, "item", "admin", check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertFalse(has_authz)

    @responses.activate
    def test_authz_wrong_perm(self):

        # Build the request
        request = self.build_request("/some/page/")

        # Build the response handler
        responses.add(
            responses.GET,
            self.authz_url_pattern,
            json={"results": [{"item": "item", "permission": "read", "email": self.user.email}]},
            status=200,
        )

        # Build the call
        has_authz = authz.has_permission(request, self.user.email, "item", "admin", check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertFalse(has_authz)

    @responses.activate
    def test_authz_case_insensitive_perm(self):

        # Build the request
        request = self.build_request("/some/page/")

        # Build the response handler
        responses.add(
            responses.GET,
            self.authz_url_pattern,
            json={"results": [{"item": "item", "permission": "admin", "email": self.user.email}]},
            status=200,
        )

        # Build the call
        has_authz = authz.has_permission(request, self.user.email, "ITEM", "ADMIN", check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertTrue(has_authz)

    @responses.activate
    def test_authz_case_insensitive_perm_1(self):

        # Build the request
        request = self.build_request("/some/page/")

        # Build the response handler
        responses.add(
            responses.GET,
            self.authz_url_pattern,
            json={"results": [{"item": "ITEM", "permission": "ADmIN", "email": self.user.email}]},
            status=200,
        )

        # Build the call
        has_authz = authz.has_permission(request, self.user.email, "item", "admin", check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertTrue(has_authz)

    @responses.activate
    def test_authz_has_perm(self):

        # Build the request
        request = self.build_request("/some/page/")

        # Build the response handler
        responses.add(
            responses.GET,
            self.authz_url_pattern,
            json={"results": [{"item": "item", "permission": "admin", "email": self.user.email}]},
            status=200,
        )

        # Build the call
        has_authz = authz.has_permission(request, self.user.email, "item", "admin", check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertTrue(has_authz)

    @responses.activate
    def test_authz_parent_has_perm(self):

        # Build the request
        request = self.build_request("/")

        # Build the response handler
        responses.add(
            responses.GET,
            self.authz_url_pattern,
            json={"results": [{"item": "parent", "permission": "admin", "email": self.user.email}]},
            status=200,
        )

        # Build the call
        has_authz = authz.has_permission(request, self.user.email, "parent.child", "admin", check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertTrue(has_authz)

    @responses.activate
    def test_authz_grandparent_has_perm(self):

        # Build the request
        request = self.build_request("/")

        # Build the response handler
        responses.add(
            responses.GET,
            self.authz_url_pattern,
            json={"results": [{"item": "parent", "permission": "admin", "email": self.user.email}]},
            status=200,
        )

        # Build the call
        has_authz = authz.has_permission(
            request, self.user.email, "parent.child.grandchild", "admin", check_parents=True
        )

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertTrue(has_authz)

    @responses.activate
    def test_authz_parent_wrong_perm(self):

        # Build the request
        request = self.build_request("/")

        # Build the response handler
        responses.add(
            responses.GET,
            self.authz_url_pattern,
            json={"results": [{"item": "parent", "permission": "read", "email": self.user.email}]},
            status=200,
        )

        # Build the call
        has_authz = authz.has_permission(request, self.user.email, "parent.child", "admin", check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertFalse(has_authz)

    @responses.activate
    def test_authz_child_perms(self):

        # Build the request
        request = self.build_request("/")

        # Build the response handler
        responses.add(
            responses.GET,
            self.authz_url_pattern,
            json={"results": [{"item": "parent.child", "permission": "admin", "email": self.user.email}]},
            status=200,
        )

        # Build the call
        has_authz = authz.has_permission(request, self.user.email, "parent", "admin", check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertFalse(has_authz)

    @responses.activate
    def test_authz_grandchild_perms(self):

        # Build the request
        request = self.build_request("/")

        # Build the response handler
        responses.add(
            responses.GET,
            self.authz_url_pattern,
            json={"results": [{"item": "parent.child.grandchild", "permission": "admin", "email": self.user.email}]},
            status=200,
        )

        # Build the call
        has_authz = authz.has_permission(request, self.user.email, "parent.child", "admin", check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertFalse(has_authz)

    @responses.activate
    def test_authz_multiple_no_perms(self):

        # Build the request
        request = self.build_request("/")

        # Build the response handler
        responses.add(
            responses.GET,
            self.authz_url_pattern,
            json={
                "results": [
                    {"item": "parent1", "permission": "admin", "email": self.user.email},
                    {"item": "parent2", "permission": "read", "email": self.user.email},
                ]
            },
            status=200,
        )

        # Build the call
        has_authz = authz.has_a_permission(
            request=request, email=self.user.email, item="parent", permissions=["admin", "read"], check_parents=True
        )

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertFalse(has_authz)

    @responses.activate
    def test_authz_multiple_perms(self):

        # Build the request
        request = self.build_request("/")

        # Build the response handler
        responses.add(
            responses.GET,
            self.authz_url_pattern,
            json={
                "results": [
                    {"item": "parent1", "permission": "admin", "email": self.user.email},
                    {"item": "parent", "permission": "read", "email": self.user.email},
                ]
            },
            status=200,
        )

        # Build the call
        has_authz = authz.has_a_permission(
            request=request, email=self.user.email, item="parent", permissions=["admin", "read"], check_parents=True
        )

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertTrue(has_authz)

    @responses.activate
    def test_authz_grandchild_multiple_perms(self):

        # Build the request
        request = self.build_request("/")

        # Build the response handler
        responses.add(
            responses.GET,
            self.authz_url_pattern,
            json={
                "results": [
                    {"item": "parent.child.grandchild", "permission": "admin", "email": self.user.email},
                    {"item": "parent.child", "permission": "read", "email": self.user.email},
                ]
            },
            status=200,
        )

        # Build the call
        has_authz = authz.has_a_permission(
            request=request,
            email=self.user.email,
            item="parent.child",
            permissions=["admin", "read"],
            check_parents=True,
        )

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertTrue(has_authz)

    @responses.activate
    def test_authz_grandchild_multiple_perms_1(self):

        # Build the request
        request = self.build_request("/")

        # Build the response handler
        responses.add(
            responses.GET,
            self.authz_url_pattern,
            json={
                "results": [
                    {"item": "parent.child.grandchild", "permission": "admin", "email": self.user.email},
                    {"item": "parent.child", "permission": "read", "email": self.user.email},
                ]
            },
            status=200,
        )

        # Build the call
        has_authz = authz.has_a_permission(
            request=request,
            email=self.user.email,
            item="parent.child",
            permissions=["admin", "write"],
            check_parents=True,
        )

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertFalse(has_authz)



import jwt
from furl import furl
import json
import base64
from jwt import PyJWKClient
from datetime import datetime, timedelta

from django.apps import apps
from django.contrib import auth as django_auth
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import PermissionDenied
from django.db.models import Q
from django.shortcuts import redirect
from django.urls import reverse
from rest_framework.authentication import BaseAuthentication
from rest_framework import exceptions

from dbmi_client.settings import dbmi_settings
from dbmi_client import authz
from dbmi_client.serializers import UserSerializer

# Get the app logger
import logging

logger = logging.getLogger(dbmi_settings.LOGGER_NAME)


# Set a key to cache JWKs under in the DBMI.AUTH0 settings
CACHED_JWKS_KEY = "__DBMI_CLIENT_CACHED_JWKS__"

# Set the name of the cookie for testing cookies
DBMI_TEST_COOKIE_NAME = "DBMI_TEST_COOKIE"


def login(request):
    """
    Builds and returns a redirect response to allow the user to login and send them to the
    supplied URL when successfully logged in. If not defined in LOGIN_REDIRECT_URL, the
    current URL in the request will be used, so the user will return to the original location.
    :param request: The original request object
    :return: The redirect response
    """
    return login_redirect(request)


def login_redirect(request, next_url=None):
    """
    Builds and returns a redirect response to allow the user to login and send them to the
    supplied URL when successfully logged in. If next_url is not passed, and
    it is not defined in LOGIN_REDIRECT_URL, the current URL in the request will be used,
    so the user will return to the original location.
    :param request: The original request object
    :param next_url: The URL users will be sent after login
    :return: Response
    """

    # Ensure the request is cleared of user state
    django_auth.logout(request)

    # Get the url
    login_url = login_redirect_url(request, next_url)

    # Just process the logout and redirect them
    response = redirect(login_url)

    # Set a test cookie
    set_login_test_cookie(request, response)

    # Do needed logout functions and return the modified response
    return response


def set_login_test_cookie(request, response):
    """
    Builds and sets a cookie on the current response to enable the authentication
    service to verify whether the user has cookies enabled or not.
    :param request: The original request object
    :type request: HttpRequest
    :param response: The to-be-returned response object
    :type response: HttpResponse
    """
    # Set a date in the future
    date = datetime.now() + timedelta(minutes=10)

    # Set a test cookie
    response.set_cookie(
        DBMI_TEST_COOKIE_NAME,
        date.strftime("%d-%m-%Y %H:%M:%S"),
        domain=dbmi_settings.JWT_COOKIE_DOMAIN,
        secure=True,
        httponly=True,
        samesite="Lax"
    )


def verify_login_test_cookie(request):
    """
    Checks for the test cookie in the current request and returns whether the
    test was successful or not.
    :param request: The original request object
    :type request: HttpRequest
    :returns: Whether the test cookie was a success or not
    :rtype: bool
    """
    # Get the test cookie
    value = request.COOKIES.get(DBMI_TEST_COOKIE_NAME)
    if not value:
        logger.info(f"Test cookie was not found, cannot proceed")
        return False

    # Compare date to ensure a recent test
    if datetime.now() >= datetime.strptime(value, "%d-%m-%Y %H:%M:%S"):
        logger.info(f"Test cookie was outdated ({value}), cannot verify cookies")
        return False

    logger.debug(f"Cookie test succeeded")
    return True


def login_redirect_url(request, next_url=None):
    """
    Builds and returns a URL that sends the user to the login, and returns them to the
    supplied URL when successfully logged in. If next_url is not passed, the current URL in the
    request will be used, so the user will return to the original location.
    :param request: The original request object
    :param next_url: The URL users will be sent after login
    :return: Response
    """
    # Check for local login enabled
    if apps.is_installed("dbmi_client.login"):

        # Use local login URL
        login_url = furl(request.build_absolute_uri(reverse("dbmi_login:authorize")))

    else:

        # Build the URL using DBMI-AuthN
        login_url = furl(dbmi_settings.AUTHN_URL)
        login_url.path.segments.extend(["login", "auth"])

    # If no next URL, determine where to dump them after logout
    if not next_url:
        if dbmi_settings.LOGIN_REDIRECT_URL:
            next_url = request.build_absolute_uri(dbmi_settings.LOGIN_REDIRECT_URL)

        else:
            next_url = request.build_absolute_uri()

    # Add next url
    login_url.query.params.add(dbmi_settings.LOGIN_REDIRECT_KEY, next_url)
    logger.debug(f"Login next URL: {next_url}")

    # Add the default client ID
    client_id = next(iter(dbmi_settings.AUTH_CLIENTS.keys()))
    login_url.query.params.add("client_id", client_id)
    logger.debug(f"Auth client ID: {client_id}")

    # Check for branding
    if dbmi_settings.AUTHN_TITLE or dbmi_settings.AUTHN_ICON_URL:

        # Add the included parameters
        branding = {}
        if dbmi_settings.AUTHN_TITLE:
            branding["title"] = dbmi_settings.AUTHN_TITLE

        if dbmi_settings.AUTHN_TITLE:
            branding["icon_url"] = dbmi_settings.AUTHN_ICON_URL

        if dbmi_settings.AUTHN_COLOR:
            branding["color"] = dbmi_settings.AUTHN_COLOR

        if dbmi_settings.AUTHN_BACKGROUND:
            branding["background"] = dbmi_settings.AUTHN_BACKGROUND

        # Encode it and pass it along
        branding_param = base64.urlsafe_b64encode(json.dumps(branding).encode("utf-8")).decode("utf-8")
        login_url.query.params.add("branding", branding_param)

    logger.debug("Login URL: {}".format(login_url.url))
    return login_url.url


def logout(request):
    """
    This will prepare the redirect to log the user out (either internally or at DBMI-AuthN)
    :param request: The Django request object
    :return: The response object that takes the user to the logout endpoint
    """
    # Call logout redirect
    return logout_redirect(request)


def logout_redirect(request, next_url=None):
    """
    This is just an alias for `logout` to support older clients
    """
    # Ensure the request is cleared of user state
    django_auth.logout(request)

    # Build the logout URL
    logout_url = logout_redirect_url(request, next_url)

    # Just process the logout and redirect them
    response = redirect(logout_url)

    # Do needed logout functions and return the modified response
    return response


def logout_redirect_url(request, next_url=None):
    """
    This will prepare the redirect URL to log the user out (either internally or at DBMI-AuthN)
    :param request: The Django request object
    :param next_url: A URL that the user should be sent to should they log back in
    :return: The response object that takes the user to the logout endpoint
    """
    # Check for local login enabled
    if apps.is_installed("dbmi_client.login"):

        # Build the URL to DBMI-Client's logout page
        logout_url = furl(request.build_absolute_uri(reverse("dbmi_login:logout")))

    else:

        # Build the URL to DBMI-AuthN's logout endpoint
        logout_url = furl(dbmi_settings.AUTHN_URL)
        logout_url.path.segments.extend(["login", "logout"])

    # If no next URL, determine where to dump them after logout
    if not next_url:
        if dbmi_settings.LOGOUT_REDIRECT_URL:
            next_url = request.build_absolute_uri(dbmi_settings.LOGOUT_REDIRECT_URL)

        else:
            next_url = request.build_absolute_uri()

    logger.debug("Logout next URL: {}".format(next_url))

    # Add next url
    logout_url.query.params.add(dbmi_settings.LOGOUT_REDIRECT_KEY, next_url)

    # Add the default client ID for the current JWT
    logout_url.query.params.add("client_id", get_jwt_client_id(request, verify=False))

    # Check for branding
    if dbmi_settings.AUTHN_TITLE or dbmi_settings.AUTHN_ICON_URL:

        # Add the included parameters
        branding = {}
        if dbmi_settings.AUTHN_TITLE:
            branding["title"] = dbmi_settings.AUTHN_TITLE

        if dbmi_settings.AUTHN_TITLE:
            branding["icon_url"] = dbmi_settings.AUTHN_ICON_URL

        # Encode it and pass it along
        branding_param = base64.urlsafe_b64encode(json.dumps(branding).encode("utf-8")).decode("utf-8")
        logout_url.query.params.add("branding", branding_param)

    logger.debug("Logout URL: {}".format(logout_url.url))
    return logout_url.url


def dbmi_http_headers(request, content_type="application/json", **kwargs):
    """
    Returns headers to be used for API calls to DBMI services in order to authenticate the caller
    :param request: The Django request
    :param content_type: The content type for the call
    :return: dict
    """

    # Get the JWT
    token = get_jwt(request)

    # Return headers
    headers = {"Authorization": "{}{}".format(dbmi_settings.JWT_HTTP_PREFIX, token), "Content-Type": content_type}

    # Add any additional parameters
    headers.update(kwargs)

    return headers


def get_jwt(request):
    """
    Takes a Django request and pulls the JWT from either cookies or HTTP headers

    :param request: The current HttpRequest object
    :type request: HttpRequest
    :returns: The JWT string
    :rtype: str
    """
    # Get the JWT token depending on request type
    if hasattr(request, "COOKIES") and request.COOKIES.get(dbmi_settings.JWT_COOKIE_NAME):
        return request.COOKIES.get(dbmi_settings.JWT_COOKIE_NAME)

    # Check if JWT in HTTP Authorization header
    elif (
        hasattr(request, "META")
        and request.META.get("HTTP_AUTHORIZATION")
        and dbmi_settings.JWT_HTTP_PREFIX in request.META.get("HTTP_AUTHORIZATION")
    ):

        # Remove prefix and return the token
        return request.META.get("HTTP_AUTHORIZATION").replace(dbmi_settings.JWT_HTTP_PREFIX, "")

    return None


def get_jwt_payload(request, verify=True):
    """
    Returns the payload from the current JWT

    :param request: The current HttpRequest object
    :type request: HttpRequest
    :param verify: Whether to verify the JWT signature first
    :type verify: bool
    :returns: The JWT payload
    :rtype: dict
    """
    # Get the JWT token depending on request type
    token = get_jwt(request)

    # Ensure we have a token
    if not token:
        return None

    # Get the payload email
    if not verify:
        try:
            return jwt.decode(
                token,
                algorithms=["RS256", "HS256"],
                options={
                    "verify_signature": False,
                },
                audience=list(dbmi_settings.AUTH_CLIENTS.keys()),
            )

        except jwt.InvalidAudienceError:
            logger.warning(f"No configuration for aud/client ID in token")
            return None
    else:
        return validate_request(request)


def get_jwt_value(request, key, verify=True):
    """
    Returns the value for the passed key from the current JWT payload

    :param request: The current HttpRequest object
    :type request: HttpRequest
    :param key: The key of the value in the JWT payload to return
    :type key: str
    :param verify: Whether to verify the JWT signature first
    :type verify: bool
    :returns: The value for the passed key
    :rtype: str
    """
    # Get the payload from above
    payload = get_jwt_payload(request, verify)
    if not payload:
        logger.debug("JWT is invalid, cannot fetch values")
        return None

    return payload.get(key)


def get_jwt_algorithm(request):
    """
    Returns the algorithm that was used to sign current JWT

    :param request: The current HttpRequest object
    :type request: HttpRequest
    :param verify: Whether to verify the JWT signature first
    :type verify: bool
    :returns: The name of the algorithm, lowercase
    :rtype: str
    """
    token = get_jwt(request)
    if not token:
        logger.debug("Could not retrieve JWT")
        return None

    return jwt.get_unverified_header(token)["alg"]



def get_jwt_username(request, verify=True):
    """
    Returns the username of the current JWT

    :param request: The current HttpRequest object
    :type request: HttpRequest
    :param verify: Whether to verify the JWT signature first
    :type verify: bool
    :returns: The current user's username
    :rtype: str
    """
    return get_jwt_value(request, "sub", verify)


def get_jwt_email(request, verify=True):
    """
    Returns the user email of the current JWT

    :param request: The current HttpRequest object
    :type request: HttpRequest
    :param verify: Whether to verify the JWT signature first
    :type verify: bool
    :returns: The current user's email
    :rtype: str
    """
    return get_jwt_value(request, "email", verify)


def get_jwt_client_id(request, verify=True):
    """
    Returns the client ID of the application used to sign current JWT

    :param request: The current HttpRequest object
    :type request: HttpRequest
    :param verify: Whether to verify the JWT signature first
    :type verify: bool
    :returns: The client ID of the JWT
    :rtype: str
    """
    return get_jwt_value(request, "aud", verify)


def get_jwt_auth_provider(request, verify=True):
    """
    Returns the provider of the application used to sign current JWT

    :param request: The current HttpRequest object
    :type request: HttpRequest
    :param verify: Whether to verify the JWT signature first
    :type verify: bool
    :returns: The provider of the current authentication provider
    :rtype: str
    """
    iss = get_jwt_value(request, "iss", verify)

    # Check URLs
    if iss.lower().contains("auth0.com"):
        return "auth0"
    elif iss.lower().contains("amazonaws.com"):
        return "cognito"


def validate_request(request):
    """
    Pulls the current cookie and verifies the JWT and
    then returns the JWT payload. Returns None
    if the JWT is invalid or missing.

    :param request: The current HttpRequest object
    :type request: HttpRequest
    :returns: The verified JWT payload
    :rtype: dict
    """
    # Extract JWT from request.
    token = get_jwt(request)

    # Get the payload
    payload = get_jwt_payload(request, verify=False)

    # Get client ID
    client_id = get_jwt_client_id(request, verify=False)

    # Ensure client ID is valid
    if client_id not in dbmi_settings.AUTH_CLIENTS:
        logger.warning(f"Client ID in token is invalid: {client_id}")
        return None

    # Fetch key based off algorithm
    algorithm = get_jwt_algorithm(request)
    if algorithm == "RS256":

        # Match the client ID to an authentication provider
        url = dbmi_settings.AUTH_CLIENTS[client_id]["JWKS_URL"]
        jwks_client = PyJWKClient(
            url,
            cache_keys=True,
            max_cached_keys=16,
            cache_jwk_set=True,
            lifespan=86400,
        )
        key = jwks_client.get_signing_key_from_jwt(token).key

    elif algorithm == "HS256":

        # Fetch key from auth configuration
        key = dbmi_settings.AUTH_CLIENTS[client_id]["CLIENT_SECRET"]

    else:
        logger.warning(f"Unsupported algorithm for token: {algorithm}")
        return None

    try:
        # Perform the validation
        payload = jwt.decode(
            token,
            key,
            algorithms=[algorithm],
            leeway=120,
            audience=client_id,
        )

        return payload

    except jwt.ExpiredSignatureError as e:
        logger.debug(
            f"JWT Expired: {e}",
        )

    except jwt.InvalidTokenError as e:
        logger.info(
            f"Invalid JWT Token: {e}",
        )

    except Exception as e:
        logger.exception(
            f"Error validating JWT: {e}",
            exc_info=True,
            extra={
                "client_id": client_id,
            },
        )

    return None

###################################################################
#
# Django Custom Authentication Backends
#
###################################################################


class DBMIAuthenticationBackend(object):
    def authenticate(self, request, token=None):
        """
        All versions of this backend follow the same flow:
        1. Check JWT is present
        2. Check JWT validity
        3. Check JWT payload for required user properties
        4. Return existing user or create new one
        5. Sync user, if already exists, with JWT payload
        """
        # Get the token
        if not token:
            token = get_jwt(request)
            if not token:
                return None

        # Validate request
        payload = validate_request(request)
        if not payload:
            return None

        # Get their email and check for their record
        email = payload.get("email")
        username = payload.get("sub")
        if not email or not username:
            logger.error("No sub or email in valid JWT: {}".format(payload))
            return None

        # The JWT is valid, now get the user object to attach to the request
        user = self._get_user_object(request)
        if not user:

            # Create the user, which will also sync
            user = self._create_user(request)

        else:

            # Sync the user to ensure we've got updated properties
            self._sync_user(request, user)

        return user

    def get_user(self, user_id):
        # Should be implemented by subclass depending on data source for user
        raise SystemError("This method should not be called")

    def has_perm(self, user, perm, obj):
        """
        Returns whether the given user has the permission or not
        """
        # This authentication backend does not manage Django-style permissions
        pass

    def has_module_perms(self, user, app_label):
        """
        Returns whether the given user has permissions on the module or not
        """
        # This authentication backend does not manage Django-style permissions
        pass

    def _get_user_object(self, request):
        """
        Accepts details from the JWT user and returns an object representing
        the request's user. If model is enabled, this will be an instance of User,
        otherwise an instance of DBMIUser
        """
        # Should be implemented by subclass depending on data source for user
        raise SystemError("This method should not be called")

    def _create_user(self, request):
        """
        Called when a user in the model does not exist. This creates the user in the
        model.
        """
        # Should be implemented by subclass depending on data source for user
        raise SystemError("This method should not be called")

    def _sync_user(self, request, user):
        """
        Called after a user is fetched/created and syncs any additional properties
        from the JWT's payload to the user object.
        """
        username = None
        email = None
        try:
            # Get the unverified payload
            payload = get_jwt_payload(request, verify=False)

            # Get properties
            username = payload["sub"].lower()
            email = payload["email"].lower()

            # Check if email or username missing
            if not user.username.lower() == username:
                logger.debug("User's username did not match JWT: {} -> {}".format(user.username, username))
                user.username = username

            if not user.email.lower() == email:
                logger.error(
                    "User's email did not match JWT: {} -> {}".format(user.email, email),
                    extra={
                        "request": request,
                        "username": user.username,
                        "email": user.email,
                        "jwt_username": username,
                        "jwt_email": email,
                    },
                )
                raise PermissionDenied

        except Exception as e:
            logger.exception(
                "User syncing error: {}".format(e),
                exc_info=True,
                extra={"request": request, "user": user, "username": username, "email": email},
            )


class DBMIJWTAuthenticationBackend(DBMIAuthenticationBackend):

    """
    Clients must have a valid JWT in the request (either in HTTP Authorization headers or in cookies).
    Users objects are an instance of DBMIJWTUser and mimic the properties and methods of Django's built-in
    contrib.auth.models.User model, but with no persistence. All properties will be valid but any attempt to
    save or link these instances to another model instance will fail.
    """

    def _get_user_object(self, request):
        """
        Accepts details from the JWT user and returns an object representing
        the request's user.
        """
        # Create an instance of the JWT user
        user = DBMIJWTUser(request)

        # Sync their profile and return them
        self._sync_user(request, user)

        return user

    def _sync_user(self, request, user):
        """
        Called after a user is fetched/created and syncs any additional properties
        from the JWT's payload to the user object.
        """
        # The user object is built from the JWT on every request so syncing is redundant and
        # not required.
        pass


class DBMIJWTAdminAuthenticationBackend(DBMIAuthenticationBackend):

    """
    Clients must have a valid JWT in the request (either in HTTP Authorization headers or in cookies).
    Users objects are an instance of DBMIJWTUser and mimic the properties and methods of Django's built-in
    contrib.auth.models.User model, but with no persistence. All properties will be valid but any attempt to
    save or link these instances to another model instance will fail. DBMI AuthZ is consulted for staff/superuser
    access and the User object is prepared as such.
    """

    def _get_user_object(self, request):
        """
        Accepts details from the JWT user and returns an object representing
        the request's user.
        """
        # Create an instance of the JWT user
        user = DBMIJWTUser(request)

        # Sync their profile and return them
        self._sync_user(request, user)

        return user

    def _sync_user(self, request, user):
        """
        Called after a user is fetched/created and syncs any additional properties
        from the JWT's payload to the user object.
        """
        # All sync admin/superuser status
        try:
            # Check if admin/superuser
            if authz.is_admin(request, user.email):
                user.is_staff = True
                user.is_superuser = True
            else:
                user.is_staff = False
                user.is_superuser = False

        except Exception as e:
            logger.exception(
                "User syncing error: {}".format(e), exc_info=True, extra={"user": user, "request": request}
            )


class DBMIModelAuthenticationBackend(DBMIAuthenticationBackend):

    """
    Clients must have a valid JWT in the request (either in HTTP Authorization headers or in cookies).
    If enabled, users are auto-created upon first attempted login with a valid JWT. The
    User model is keyed by the username and email contained in the JWT. Profile and groups are synced
    from the JWT upon each login.
    """

    def user_can_authenticate(self, user):
        """
        Reject users with is_active=False. Custom user models that don't have
        that attribute are allowed.
        """
        is_active = getattr(user, "is_active", None)
        return is_active or is_active is None

    def get_user(self, user_id):
        UserModel = django_auth.get_user_model()
        try:
            user = UserModel._default_manager.get(pk=user_id)

        except UserModel.DoesNotExist:
            return None

        return user if self.user_can_authenticate(user) else None

    def _get_user_object(self, request):
        """
        Accepts details from the JWT user and returns an object representing
        the request's user. If model is enabled, this will be an instance of User,
        otherwise an instance of DBMIUser
        """
        # Get username and email
        username = get_jwt_username(request, verify=False)
        email = get_jwt_email(request, verify=False)

        # Fetch the current User model
        UserModel = django_auth.get_user_model()

        # Find the user
        users = UserModel.objects.filter(Q(username=username) | Q(email=email)).order_by("-date_joined")
        if not users:
            logger.debug("User does not yet exist: {} : {}".format(username, email))
            return None

        # Get first user
        user = users.first()

        # Check for duplicate users (on email)
        if len(users) > 1:
            logger.error(
                "Found {} users for {} : {}".format(len(users), email, ",".join([u.username for u in users])),
                extra={"request": request, "users": json.dumps(UserSerializer(users, many=True).data)},
            )
            logger.debug("Returning oldest user: {} : {}".format(user.username, user.email))

        else:
            logger.debug("Found user: {} : {}".format(username, email))

        return user

    def _create_user(self, request):
        """
        Called when a user in the model does not exist. This creates the user in the
        model.
        """
        # Check if autocreate is enabled
        if not dbmi_settings.USER_MODEL_AUTOCREATE:
            logger.debug("User autocreate is disabled, boot the current user")
            raise PermissionDenied

        # Get username and email
        username = get_jwt_username(request, verify=False)
        email = get_jwt_email(request, verify=False)

        # Create them
        UserModel = django_auth.get_user_model()
        user = UserModel(username=username, email=email)
        user.set_unusable_password()
        logger.debug("Created user: {}:{}".format(username, email))

        # Sync them up
        self._sync_user(request, user)

        return user

    def _set_superuser(self, request, user):
        """
        This method allows backends to determine whether or not an admin user
        should be granted Django superuser status.

        :param request: The current request
        :type request: HttpRequest
        :param user: The current user
        :type user: User
        """
        return False

    def _set_staff(self, request, user):
        """
        This method allows backends to determine whether or not an admin user
        should be granted Django superuser status.

        :param request: The current request
        :type request: HttpRequest
        :param user: The current user
        :type user: User
        """
        return False

    def _sync_user(self, request, user):
        """
        Called after a user is fetched/created and syncs any additional properties
        from the JWT's payload to the user object.
        """
        username = None
        email = None
        try:
            # Get the unverified payload
            payload = get_jwt_payload(request, verify=False)

            # Get properties
            username = payload["sub"].lower()
            email = payload["email"].lower()

            # Check if email or username missing
            if not user.username.lower() == username:
                logger.debug("User's username did not match JWT: {} -> {}".format(user.username, username))
                user.username = username

            if not user.email.lower() == email:
                logger.error(
                    "User's email did not match JWT: {} -> {}".format(user.email, email),
                    extra={
                        "request": request,
                        "username": user.username,
                        "email": user.email,
                        "jwt_username": username,
                        "jwt_email": email,
                    },
                )
                raise PermissionDenied

            # Check if superuser/staff
            if self._set_staff(request, user):
                user.is_staff = True

            if self._set_superuser(request, user):
                user.is_superuser = True

            # Save
            user.save()

        except Exception as e:
            logger.exception(
                "User syncing error: {}".format(e),
                exc_info=True,
                extra={"request": request, "user": user, "username": username, "email": email},
            )


class DBMIUsersModelAuthenticationBackend(DBMIModelAuthenticationBackend):

    """
    Clients must have a valid JWT in the request (either in HTTP Authorization headers or in cookies).
    Use this authentication backend for sites that are accessible to all users as well as administrator users.
    User model is keyed by the username and email contained in the JWT. `is_staff` and `is_superuser` flags
    are automatically set and synced on appropriate users. Every active user with admin authorization on this site
    will have complete access to everything.
    """

    def _sync_user(self, request, user):
        """
        Called after a user is fetched/created and syncs any additional properties
        from the JWT's payload to the user object. Set staff and superuser flags
        if authorizations are valid.
        """
        # Do normal sync first
        super(DBMIModelAuthenticationBackend, self)._sync_user(request, user)

        is_admin = False
        try:
            # Check if admin
            is_admin = authz.is_admin(request, user.email)
            if is_admin:
                logger.debug(f"User: {user.email} has been granted admin/superuser privileges")

            # Ensure the model is updated
            user.is_staff = is_admin
            user.is_superuser = is_admin
            user.save()

        except Exception as e:
            logger.exception(
                "Superuser syncing error: {}".format(e),
                exc_info=True,
                extra={
                    "request": request,
                    "user": user,
                    "username": user.username,
                    "email": user.email,
                    "is_admin": is_admin,
                },
            )


class DBMIRestrictedModelAuthenticationBackend(DBMIModelAuthenticationBackend):

    """
    Clients must have a valid JWT in the request (either in HTTP Authorization headers or in cookies) as
    well as authorization, either through JWT claims or as a permission in the DBMI AuthZ service.
    A method performs an inspection of any proposed users' authorizations and permissions and determines
    if they should be created within the model or not. The default implementation requires proposed
    users to have any existing permissions on the current application.

    Use this authentication backend for sites that are only accessible to users with specific
    permissions/roles/groups. User model is keyed by the username and email contained in the JWT.
    Profile and groups are synced from the JWT upon each login.
    """

    def _should_create_user(self, request, email):
        """
        This method inspects the proposed user and returns whether they should be created or not.
        Typically, before a user is rejected due to a missing permission, their record will have
        already been created in the Django model. This method allows subclassing backends to do
        a check of the proposed user before the creation step, thus avoiding User entries that
        would never be able to log in anyways. The default implementation checks for any existing
        permissions on the current application.

        :param request: The current request
        :type request: HttpRequest
        :param email: The email of the requesting user
        :type email: str
        :return: Whether the user should be created or not
        :rtype: bool
    """
        return len(authz.get_permissions(request, email, item=dbmi_settings.CLIENT, children=True)) > 1

    def _create_user(self, request):
        """
        This middleware performs exactly like its superclass, with the exception of checking
        an authenticated user's authorizations before creating them in the model. This would
        be used for sites where only admins/superusers/staff should have access.
        """
        # Get username and email
        username = get_jwt_username(request, verify=False)
        email = get_jwt_email(request, verify=False)

        # Before we create a user, we must ensure they have authorizations
        if not self._should_create_user(request, email):
            logger.debug('User {}/{} did not pass restriction test, access denied'.format(username, email))
            raise PermissionDenied

        # Create them
        UserModel = django_auth.get_user_model()
        user = UserModel(username=username, email=email)
        user.set_unusable_password()
        logger.debug("Created user: {}:{}".format(username, email))

        # Sync them up
        self._sync_user(request, user, is_admin=True)

        return user

    def _sync_user(self, request, user, is_admin=None):
        """
        Called after a user is fetched/created and syncs any additional properties
        from the JWT's payload to the user object. This method is extended to
        add the flag determining admin authorization to prevent multiple
        requests to the authz server.
        """
        # Do normal sync first
        super(DBMIRestrictedModelAuthenticationBackend, self)._sync_user(request, user)


class DBMIAdminModelAuthenticationBackend(DBMIModelAuthenticationBackend):

    """
    Clients must have a valid JWT in the request (either in HTTP Authorization headers or in cookies) as
    well as authorization, either through JWT claims or as a permission in the DBMI AuthZ service.
    A method performs an inspection of any proposed users' authorizations and permissions and determines
    if they should be created within the model or not. The default implementation requires proposed
    users to have an 'admin' permission set on the current app, but can be overidden to check for
    other authorizations.

    Use this authentication backend for sites that are only accessible to users with specific
    permissions/roles/groups. User model is keyed by the username and email contained in the JWT.
    Profile and groups are synced from the JWT upon each login.
    """

    def _should_create_user(self, request, email):
        """
        This method inspects the proposed user and returns whether they should be created or not.
        Typically, before a user is rejected due to a missing permission, their record will have
        already been created in the Django model. This method allows subclassing backends to do
        a check of the proposed user before the creation step, thus avoiding User entries that
        would never be able to log in anyways. The default implementation checks for a single
        'admin' permission on the current application.

        :param request: The current request
        :type request: HttpRequest
        :param email: The email of the requesting user
        :type email: str
        :return: Whether the user should be created or not
        :rtype: bool
        """
        return authz.is_admin(request, email)


class DBMISuperuserModelAuthenticationBackend(DBMIAdminModelAuthenticationBackend):

    """
    Clients must have a valid JWT in the request (either in HTTP Authorization headers or in cookies) as
    well as admin authorization, either through JWT claims or as a permission in the DBMI AuthZ service.
    Use this authentication backend for sites that are only accessible to admins and no other users.
    User model is keyed by the username and email contained in the JWT. `is_staff` and `is_superuser` flags
    are automatically set and synced on users. Every active user with authorization on this site
    will have complete access to everything.
    """

    def _sync_user(self, request, user, is_admin=None):
        """
        Called after a user is fetched/created and syncs any additional properties
        from the JWT's payload to the user object. Set staff and superuser flags
        if authorizations are valid.
        """
        # Do normal sync first
        super(DBMISuperuserModelAuthenticationBackend, self)._sync_user(request, user)

        try:
            # Check if admin
            if is_admin is None:
                is_admin = authz.is_admin(request, user.email)

            # Ensure the model is updated
            user.is_staff = is_admin
            user.is_superuser = is_admin
            user.save()

            # If not admin (indicates they used to be), save and raise exception
            if not is_admin:
                logger.debug("User was superuser, but is now missing authz, booting them: {}".format(user.username))
                raise PermissionDenied

        except Exception as e:
            logger.exception(
                "Superuser syncing error: {}".format(e),
                exc_info=True,
                extra={
                    "request": request,
                    "user": user,
                    "username": user.username,
                    "email": user.email,
                    "is_admin": is_admin,
                },
            )

            logger.debug("Encountered an issue and could not check admin/superuser status: defaulting to access denied")
            raise PermissionDenied


###################################################################
#
# Django user object for user-less databases
#
###################################################################


class DBMIJWTUser(AnonymousUser):
    """
    This class represents a JWT user for an application that does not persist those
    users to the store. It provides a request.user object to access user properties
    but will throw exceptions if ever used in the context of database operations.
    """

    email = None
    username = None
    id = None
    pk = None
    is_active = True
    is_staff = False
    is_superuser = False
    jwt = None

    def __init__(self, request):

        # Get the payload
        payload = get_jwt_payload(request, verify=False)
        self.jwt = get_jwt(request)

        # Set properties
        self.username = payload.get("sub")
        self.id = payload.get("sub")
        self.email = payload.get("email")

    def __str__(self):
        return self.id

    def __eq__(self, other):
        return hasattr(other, "id") and self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def save(self):
        raise NotImplementedError("Django doesn't provide a DB representation for DBMIJWTUser.")

    def delete(self):
        raise NotImplementedError("Django doesn't provide a DB representation for DBMIJWTUser.")

    def set_password(self, raw_password):
        raise NotImplementedError("Django doesn't provide a DB representation for DBMIJWTUser.")

    def check_password(self, raw_password):
        raise NotImplementedError("Django doesn't provide a DB representation for DBMIJWTUser.")

    @property
    def is_anonymous(self):
        return False

    @property
    def is_authenticated(self):
        return True

    def get_username(self):
        return self.username

    def get_group_permissions(self, obj=None):
        """
        Return a list of permission strings that this user has through their
        groups. Query all available auth backends. If an object is passed in,
        return only permissions matching this object.
        """
        permissions = set()
        return permissions

    def get_all_permissions(self, obj=None):
        # Get permissions from AuthZ
        return authz.get_permissions(self.jwt, self.email, obj)

    def has_perm(self, perm, obj=None):
        """
        Return True if the user has the specified permission. Query all
        available auth backends, but return immediately if any backend returns
        True. Thus, a user who has permission from a single auth backend is
        assumed to have permission in general. If an object is provided, check
        permissions for that object.
        """
        # Active superusers have all permissions.
        if self.is_active and self.is_superuser:
            return True

        # Otherwise we need to check the backends.
        return _user_has_perm(self, perm, obj)

    def has_perms(self, perm_list, obj=None):
        """
        Return True if the user has each of the specified permissions. If
        object is passed, check if the user has all required perms for it.
        """
        return all(self.has_perm(perm, obj) for perm in perm_list)

    def has_module_perms(self, app_label):
        """
        Return True if the user has any permissions in the given app label.
        Use similar logic as has_perm(), above.
        """
        # Active superusers have all permissions.
        if self.is_active and self.is_superuser:
            return True

        return _user_has_module_perms(self, app_label)


def _user_has_perm(user, perm, obj):
    """
    A backend can raise `PermissionDenied` to short-circuit permission checking.
    """
    for backend in django_auth.get_backends():
        if not hasattr(backend, "has_perm"):
            continue
        try:
            if backend.has_perm(user, perm, obj):
                return True
        except PermissionDenied:
            return False
    return False


def _user_has_module_perms(user, app_label):
    """
    A backend can raise `PermissionDenied` to short-circuit permission checking.
    """
    for backend in django_auth.get_backends():
        if not hasattr(backend, "has_module_perms"):
            continue
        try:
            if backend.has_module_perms(user, app_label):
                return True
        except PermissionDenied:
            return False
    return False


###################################################################
#
# Django Rest Framework (DRF) Custom Authentication/Authorization
#
###################################################################


class DBMIUser(BaseAuthentication):
    """
    Authentication method for DBMI API methods
    """

    def authenticate(self, request):

        # Get the JWT
        token = get_jwt(request)
        if not token:
            raise exceptions.NotAuthenticated

        # User has a valid JWT from SciAuth
        payload = validate_request(request)
        if not payload:
            raise exceptions.AuthenticationFailed

        # Return the user's email to attach to the request object (request.user)
        # Also, return the authz dictionary contained in the JWT claims, if present (request.auth)
        return payload.get("email"), payload.get(dbmi_settings.JWT_AUTHZ_NAMESPACE)


class DBMIModelUser(BaseAuthentication):
    """
    Authentication method for DBMI API methods
    """

    def authenticate(self, request):

        # Get the JWT
        token = get_jwt(request)
        if not token:
            raise exceptions.NotAuthenticated

        # Call the standard Django authenticate method, that will in
        # turn call DBMIModelAuthenticationBackend.authenticate
        user = django_auth.authenticate(request, token=token)
        if not user:
            raise exceptions.AuthenticationFailed

        # Check if JWT contains AuthZ
        auth = None
        if dbmi_settings.JWT_AUTHZ_NAMESPACE:

            # User has a valid JWT from SciAuth
            auth = get_jwt_payload(request, verify=False).get("JWT_AUTHZ_NAMESPACE")

        # Return the user's email to attach to the request object (request.user)
        # Also, return the authz dictionary contained in the JWT claims, if present (request.auth)
        return user, auth



from furl import furl
import requests
import json
import base64

from dbmi_client import authn
from dbmi_client.settings import dbmi_settings


# Get the app logger
import logging

logger = logging.getLogger(dbmi_settings.LOGGER_NAME)


def create_dbmi_user(request, **profile):
    logger.debug("Creating DBMI user")

    # Get the JWT
    email = authn.get_jwt_email(request, verify=False)

    # Update kwargs
    profile["email"] = email

    # Build the URL (needs trailing slash)
    url = furl(dbmi_settings.REG_URL)
    url.path.segments.extend(["api", "register", ""])

    response = requests.post(url.url, headers=authn.dbmi_http_headers(request), data=json.dumps(profile))
    if not response.ok:
        logger.error("Create user response: {}".format(response.content))

    return response.json()


def get_dbmi_user(request, email=None):
    logger.debug("Get DBMI user")

    # Get the JWT
    if not email:
        email = authn.get_jwt_email(request, verify=False)

    # Build the URL (needs trailing slash)
    url = furl(dbmi_settings.REG_URL)
    url.path.segments.extend(["api", "register", ""])

    # Add email
    url.query.params.add("email", email)

    # Requests for profiles are limited to the profile for the requesting user
    response = requests.get(url.url, headers=authn.dbmi_http_headers(request))
    if not response.ok:
        logger.error("Get user response: {}".format(response.content))

    # Return the profile
    profiles = response.json()["results"]
    return next(iter(profiles), None)


def update_dbmi_user(request, **profile):
    logger.debug("Update DBMI user")

    # Get the JWT
    email = authn.get_jwt_email(request, verify=False)

    # Get their profile first
    reg_profile = get_dbmi_user(request=request, email=email)
    if not reg_profile:

        # Ensure email is in their profile
        if "email" not in profile:
            profile["email"] = email

        # Create the profile
        return create_dbmi_user(request, **profile)

    else:
        # Build the URL (needs trailing slash)
        url = furl(dbmi_settings.REG_URL)
        url.path.segments.extend(["api", "register", reg_profile["id"], ""])

        response = requests.put(url.url, headers=authn.dbmi_http_headers(request), data=json.dumps(profile))
        if not response.ok:
            logger.error("Update user response: {}".format(response.content))

        return response.json()


def send_email_confirmation(request, success_url, title=None, icon=None, subject=None):
    logger.debug("Sending confirmation email")

    # Build the URL (needs trailing slash)
    url = furl(dbmi_settings.REG_URL)
    url.path.segments.extend(["api", "register", "send_confirmation_email", ""])

    # Add extra data to define look and feel of email, if passed
    branding = {}
    if title:
        branding["title"] = title
    if icon:
        branding["icon"] = icon
    if subject:
        branding["subject"] = subject

    # Set data for request
    data = {
        "success_url": success_url,
    }

    # Check for branding
    if branding:
        data["branding"] = base64.b64encode(json.dumps(branding).encode()).decode()

    # Make the call
    response = requests.post(url.url, headers=authn.dbmi_http_headers(request), data=json.dumps(data))
    if not response.ok:
        logger.error("Confirmation email response: {}".format(response.content))

    return response.ok


def check_email_confirmation(request):
    logger.debug("Checking email confirmation")

    # Build the URL (needs trailing slash)
    url = furl(dbmi_settings.REG_URL)
    url.path.segments.extend(["api", "register", ""])
    url.query.params.add("email", authn.get_jwt_email(request, verify=False))

    # Make the call
    response = requests.get(url.url, headers=authn.dbmi_http_headers(request))
    if not response.ok:
        logger.error("Confirmation email response: {}".format(response.content))
        return None

    try:
        # Parse the profile for the status
        email_status = response.json()["results"][0]["email_confirmed"]
        logger.debug("Email confirmation status: {}".format(email_status))

        return email_status

    except (KeyError, IndexError) as e:
        logger.error("Failed parsing profile: {}".format(e))

    return None



from django.urls import re_path

from dbmi_client.login import views

app_name = "dbmi_login"

urlpatterns = [
    re_path(r"^auth(?:orize)?/?$", views.authorize, name="authorize"),
    re_path(r"^callback/?$", views.callback, name="callback"),
    re_path(r"^logout/?$", views.logout, name="logout"),
    re_path(r"^token/?$", views.token, name="token"),
    re_path(r"^$", views.token, name="token"),
]



#!/usr/bin/env python
"""
This script is a trick to setup a fake Django environment, since this reusable
app will be developed and tested outside any specifiv Django project.
Via ``settings.configure`` you will be able to set all necessary settings
for your app and run the tests as if you were calling ``./manage.py test``.
"""
import sys
import pytest
from django.conf import settings

EXTERNAL_APPS = [
    "django.contrib.admin",
    "django.contrib.admindocs",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.messages",
    "django.contrib.sessions",
    "django.contrib.staticfiles",
    "django.contrib.sitemaps",
    "django.contrib.sites",
]
INTERNAL_APPS = [
    "dbmi_client",
]
INSTALLED_APPS = EXTERNAL_APPS + INTERNAL_APPS

if not settings.configured:
    settings.configure(
        SECRET_KEY="*4n1!z0%@w-e&u9c-kpyqof=-nxvx^v2m000#gf9vewm3s+_v)",
        DATABASES={
            "default": {
                "ENGINE": "django.db.backends.sqlite3",
                "NAME": ":memory:",
            }
        },
        INSTALLED_APPS=INSTALLED_APPS,
        DBMI_CLIENT_CONFIG={
            "CLIENT": "DBMI-Client",
            "ENVIRONMENT": "prod",
            "AUTH_CLIENTS": {"someauthclientid":{"JWKS_URL":"https://somejwksurl.com/","PROVIDER":"cognito"}},
        },
    )


def main(*test_args):
    sys.exit(pytest.main(["-x"]))


if __name__ == "__main__":
    main(*sys.argv[1:])



import requests
import json
from furl import furl

from django.core.mail import send_mail
from dbmi_client.settings import dbmi_settings

# Get the app logger
import logging
logger = logging.getLogger(dbmi_settings.LOGGER_NAME)


class Support:
    """
    This class manages support requests with Jira Support Desk. Utility methods
    are provided designed to submit requests on behalf of users.
    """
    @classmethod
    def _headers(cls):
        """
        Returns headers used to qualify requests to Jira's API
        """
        return {
            'X-ExperimentalApi': 'true'
        }

    @classmethod
    def _auth(cls):
        """
        Returns the authentication object used to authenticate calls to Jira's
        API.
        """
        if not dbmi_settings.JIRA_USERNAME or not dbmi_settings.JIRA_TOKEN:
            raise SystemError("Cannot use Support without configured Jira credentials")

        return (dbmi_settings.JIRA_USERNAME, dbmi_settings.JIRA_TOKEN)

    @classmethod
    def _get_paged_results(cls, url, filter=None, limit=None):
        """
        Accepts a URL and returns all of the paged results, combined. Only
        simple top-level key-value matching is supported.

        :param url: The URL to fetch from
        :type url: str
        :param filter: A lambda to filter results on
        :type filter: lambda, defaults to None
        :param limit: A number to limit results too
        :type limit: int, defaults to None
        :return: A list of all objects returned
        :rtype: list
        """
        content = None
        try:
            # Collect results
            objects = []

            # Page results
            next_url = url
            while next_url is not None:

                # Pull a page
                response = requests.get(next_url, headers=cls._headers(), auth=cls._auth())
                content = response.content
                response.raise_for_status()

                # Check for filter and apply it
                if filter and filter(object):

                    # Append the matched object
                    objects.append(object)

                else:
                    # Parse response
                    objects.extend(response.json().get("values", []))

                # Check limit
                if limit and len(objects) >= limit:
                    return objects

                # Check for next URL
                next_url = response.json().get("_links", {}).get("next", None)

            return objects

        except Exception as e:
            logger.exception(
                f"Support: Error querying Jira: {e}",
                exc_info=True,
                extra={
                    "url": url,
                    "content": content,
                }
            )

    @classmethod
    def get_organization(cls, name=None):
        """
        Returns a list of current organizations for the service desk. If passed,
        filters by name.

        :param email: The email of the customer to find, defaults to None
        :type email: str, optional
        """
        # Make the request
        url = furl(f"https://{dbmi_settings.JIRA_ORGANIZATION}.atlassian.net")
        url.path.segments.extend(["rest", "servicedeskapi", "organization"])

        organizations = cls._get_paged_results(url.url)

        # Check if searching
        if name:

            # Find them
            return next((o for o in organizations if o["name"].lower() == name.lower()), None)

        else:

            return organizations

    @classmethod
    def get_organization_id(cls, name):
        """
        Returns a list of current organizations for the service desk. If passed,
        filters by name.

        :param email: The email of the customer to find, defaults to None
        :type email: str, optional
        """
        # Return it
        return cls.get_organization(name=name)["id"]

    @classmethod
    def get_customers(cls, email=None):
        """
        Returns a list of current customers for the service desk. If passed,
        filters by email address.

        :param email: The email of the customer to find, defaults to None
        :type email: str, optional
        """
        # Make the request
        url = furl(f"https://{dbmi_settings.JIRA_ORGANIZATION}.atlassian.net")
        url.path.segments.extend(["rest", "servicedeskapi", "servicedesk", dbmi_settings.JIRA_SERVICE_DESK, "customer"])

        customers = cls._get_paged_results(url.url)

        # Check if searching
        if email:

            # Find them
            return next((c for c in customers if c["emailAddress"].lower() == email.lower()), None)

        else:

            return customers

    @classmethod
    def get_customer_id(cls, email):
        """
        Returns the ID of the user for the passed email, or None if the
        customer does not exist.

        :param email: The email to search on
        :type email: str
        """
        # Get customers
        customer = cls.get_customers(email=email)
        if customer:

            return customer["accountId"]

        return None

    @classmethod
    def get_service_desks(cls):
        """
        Gets the service desks object from Jira
        """
        # Make the request
        url = furl(f"https://{dbmi_settings.JIRA_ORGANIZATION}.atlassian.net")
        url.path.segments.extend(["rest", "servicedeskapi", "servicedesk"])

        return cls._get_paged_results(url.url)

    @classmethod
    def get_service_desk_id(cls):
        """
        Gets the service desk ID from Jira
        """
        return next((s["id"] for s in cls.get_service_desks() if s["projectKey"] == dbmi_settings.JIRA_SERVICE_DESK), None)

    @classmethod
    def get_request_types(cls):
        """
        Gets the request types for the configured service desk from Jira
        """
        # Make the request
        url = furl(f"https://{dbmi_settings.JIRA_ORGANIZATION}.atlassian.net")
        url.path.segments.extend(["rest", "servicedeskapi", "servicedesk", dbmi_settings.JIRA_SERVICE_DESK, "requesttype"])

        return cls._get_paged_results(url.url)

    @classmethod
    def create_customer(cls, email, name):
        """
        Creates a customer in Jira Support Desk and returns their ID.

        :param email: The email of the user to create
        :type email: str
        :param name: The name of the customer to create
        :type name: str
        :return: A tuple of operation success and response data
        :rtype: boolean, object
        """
        content = None
        try:
            # Make the request
            url = furl(f"https://{dbmi_settings.JIRA_ORGANIZATION}.atlassian.net")
            url.path.segments.extend(["rest", "servicedeskapi", "customer"])

            # Set the data
            data = {
                "email": email,
                "fullName": name,
            }

            # Pull a page
            response = requests.post(url.url, json=data, headers=cls._headers(), auth=cls._auth())
            content = response.content
            response.raise_for_status()

            # Return ID
            return response.ok, response.json()

        except Exception as e:
            logger.exception(
                f"Support: Error creating Jira customer: {e}",
                exc_info=True,
                extra={
                    'email': email,
                    'response': response,
                }
            )

        return False, content

    @classmethod
    def add_customer_to_service_desk(cls, customer_id):
        """
        Adds a customer in Jira Support Desk to the configured
        service desk.

        :param customer_id: The ID of the user to add
        :type customer_id: str
        :return: A tuple of operation success and response data
        :rtype: boolean, object
        """
        content = None
        try:
            # Make the request /rest/servicedeskapi/servicedesk/{serviceDeskId}/customer
            url = furl(f"https://{dbmi_settings.JIRA_ORGANIZATION}.atlassian.net")
            url.path.segments.extend(["rest", "servicedeskapi", "servicedesk", dbmi_settings.JIRA_SERVICE_DESK, "customer"])

            # Set the data
            data = {"usernames":[customer_id]}

            # Pull a page
            response = requests.post(url.url, json=data, headers=cls._headers(), auth=cls._auth())
            content = response.content
            response.raise_for_status()

            # Return ID
            return response.ok, response.json()

        except Exception as e:
            logger.exception(
                f"Support: Error adding customer to service desk: {e}",
                exc_info=True,
                extra={
                    'customer_id': customer_id,
                    'response': response,
                }
            )

        return False, content

    @classmethod
    def add_customer_to_organization(cls, customer_id, organization_id):
        """
        Adds a customer in Jira Support Desk to the passed organization

        :param customer_id: The ID of the user to add
        :type customer_id: str
        :param organization_id: The ID of the organization
        :type organization_id: str
        :return: A tuple of operation success and response data
        :rtype: boolean, object
        """
        content = None
        try:
            # Make the request
            url = furl(f"https://{dbmi_settings.JIRA_ORGANIZATION}.atlassian.net")
            url.path.segments.extend(["rest", "servicedeskapi", "organization", organization_id, "user"])

            # Set the data
            data = {"usernames":[customer_id]}

            # Pull a page
            response = requests.post(url.url, json=data, headers=cls._headers(), auth=cls._auth())
            content = response.content
            response.raise_for_status()

            # Return ID
            return response.ok, response.json()

        except Exception as e:
            logger.exception(
                f"Support: Error adding customer to organization: {e}",
                exc_info=True,
                extra={
                    'customer_id': customer_id,
                    'organization_id': organization_id,
                    'response': response,
                }
            )

        return False, content

    @classmethod
    def create_request(cls, customer, request_type_id, subject, request, labels=None):
        """
        Creates a request for the passed customer.

        :param customer: The customer ID or email to create the request on behalf of
        :type customer: str
        :param request_type_id: The request type ID
        :type request_type_id: str
        :param subject: The subject of the request
        :type subject: str
        :param request: The body of the request
        :type request: str
        :return: Whether the request succeeded and the object
        :rtype: bool, dict
        """
        content = None
        try:
            # Get the service desk ID
            service_desk_id = cls.get_service_desk_id()

            # Make the request
            url = furl(f"https://{dbmi_settings.JIRA_ORGANIZATION}.atlassian.net")
            url.path.segments.extend(["rest", "servicedeskapi", "request"])

            # Set the data
            data = {
                "serviceDeskId": service_desk_id,
                "requestTypeId": request_type_id,
                "requestFieldValues": {
                    "summary": subject,
                    "description": request
                },
                "raiseOnBehalfOf": customer,
            }

            # Check for labels
            if labels and type(labels) is list:

                # Add them
                data["requestFieldValues"]["labels"] = labels

            # Pull a page
            response = requests.post(url.url, json=data, headers=cls._headers(), auth=cls._auth())
            content = response.content
            response.raise_for_status()

            # Return ID
            return response.ok, response.json()

        except Exception as e:
            logger.exception(
                f"Support: Error creating Jira request: {e}",
                exc_info=True,
                extra={
                    'customer': customer,
                    'request_type': request_type_id,
                    'labels': labels,
                    'response': response,
                }
            )

        return False, content

    @classmethod
    def email_request(cls, email, subject, message):
        """
        Emails a request on behalf of the user/customer

        :param email: The customer email to create the request on behalf of
        :type email: str
        :param subject: The subject of the request
        :type subject: str
        :param message: The body of the request
        :type message: str
        :return: Whether the request succeeded and the object
        :rtype: bool
        """
        # Check configs
        if not dbmi_settings.JIRA_SERVICE_DESK_EMAIL:
            raise SystemError(f"Cannot email without configured Jira service desk email")

        try:
            # Send it
            success = send_mail(
                subject=subject,
                message=message,
                from_email=email,
                recipient_list=[dbmi_settings.JIRA_SERVICE_DESK_EMAIL],
                fail_silently=False,
            )

            return success

        except Exception as e:
            logger.exception(
                f"Support: Error emailing Jira: {e}",
                exc_info=True,
            )

        return False



import base64
from furl import furl
import secrets
from urllib.parse import urlencode
import pytz
from datetime import datetime
from cryptography.fernet import MultiFernet, Fernet
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.hkdf import HKDF
from cryptography.hazmat.backends import default_backend
from cryptography.fernet import InvalidToken
from django.conf import settings
from django.utils.encoding import force_bytes

from django.conf import settings
from django.core.exceptions import SuspiciousOperation
from django.shortcuts import render, redirect, reverse
from django.http import QueryDict
from dbmi_client.settings import dbmi_settings
from dbmi_client.authn import validate_request, get_jwt, get_jwt_client_id, get_jwt_payload
from django.contrib import auth as django_auth
from dbmi_client.provider import ProviderFactory
from dbmi_client.auth import dbmi_user

# Get the logger
import logging
logger = logging.getLogger(dbmi_settings.LOGGER_NAME)

# Set a name for staching state during authentication
DBMI_AUTH_STATE_COOKIE_NAME = "DBMI_AUTH_STATE"
DBMI_AUTH_QUERY_COOKIE_NAME = "DBMI_AUTH_QUERY"
DBMI_AUTH_QUERY_CLIENT_ID_KEY = "client_id"
DBMI_AUTH_QUERY_NEXT_KEY = dbmi_settings.LOGIN_REDIRECT_KEY
DBMI_AUTH_QUERY_BRANDING_KEY = "branding"
DBMI_AUTH_CALLBACK_QUERY_KEY = "query"
DBMI_AUTH_STATE_KEY = "state"
DBMI_AUTH_LOGOUT_COOKIE_NAME = "DBMI_AUTH_LOGOUT"


def derive_fernet_key(input_key):
    """
    Derive a 32-bit b64-encoded Fernet key from arbitrary input key.

    :param input_key: The key to convert to Fernet-compatible key
    :type input_key: str
    :returns: A Fernet-compatible key
    :rtype: bytes
    """
    backend = default_backend()
    info = b'dbmi-client'
    salt = b'dbmi-client-hkdf-salt'
    hkdf = HKDF(
        algorithm=hashes.SHA256(),
        length=32,
        salt=salt,
        info=info,
        backend=backend,
    )
    return base64.urlsafe_b64encode(hkdf.derive(force_bytes(input_key)))


@dbmi_user
def token(request):
    """
    This view is merely a landing page for already logged in users. It will
    display the value of their current JWT for manually signing requests and
    let them know when it expires.

    :param request: The current HTTP request object
    :type request: HttpRequest
    """

    # Get JWT details
    payload = get_jwt_payload(request, verify=True)
    expiration_datetime = datetime.utcfromtimestamp(payload['exp']).replace(tzinfo=pytz.utc)
    expiration_et_datetime = expiration_datetime.astimezone(pytz.timezone('US/Eastern'))

    # Set their JWT
    context = {
        'title': dbmi_settings.AUTHN_TITLE,
        'jwt': get_jwt(request),
        'jwt_expiration': expiration_et_datetime.strftime("%A %B %-d, %Y at %I:%M:%S %p ET"),
    }

    return render(request, template_name="dbmi_client/login/token.html", context=context)


def authorize(request):
    """
    This is the authorize endpoint and it will verify a user is logged in and
    redirect them accordingly or proceed with the OAuth2 process to log the user
    in with the given auth provider.

    :param request: The current HTTP request object
    :type request: HttpRequest
    """
    logger.debug("Checking auth")

    # Check for an existing valid DBMI JWT
    if validate_request(request):

        # Determine where to send them
        if hasattr(dbmi_settings, "LOGIN_REDIRECT_KEY") and request.GET.get(dbmi_settings.LOGIN_REDIRECT_KEY):
            redirect_url = request.GET.get(dbmi_settings.LOGIN_REDIRECT_KEY)

        elif hasattr(dbmi_settings, "LOGIN_REDIRECT_URL"):
            redirect_url = dbmi_settings.LOGIN_REDIRECT_URL

        else:
            redirect_url = request.build_absolute_uri(reverse("dbmi_login:jwt"))

        # Log their next destination
        logger.debug(f"Logged in, forward user to: {redirect_url}")

        return redirect(redirect_url)

    # Get the client ID or the default client ID which is first in the settings
    client_id = request.GET.get(DBMI_AUTH_QUERY_CLIENT_ID_KEY, next(iter(dbmi_settings.AUTH_CLIENTS.keys())))

    # Build a URL with the root URI
    callback_url = furl(request.build_absolute_uri(reverse("dbmi_login:callback")))
    logger.debug(f"Callback URL: {callback_url.url}")

    # Intialize the authentication backend
    provider = ProviderFactory.create(client_id, callback_url.url)
    logger.debug(f"Provider '{provider.identifier}' matched to client ID '{client_id}'")

    # Build an encoded querystring to pass along as state
    state = {}
    logger.debug(f"Passed parameters: {','.join(list(request.GET.keys()))}")

    # Add redirect URL
    if request.GET.get(DBMI_AUTH_QUERY_NEXT_KEY):

        # Add it
        state[DBMI_AUTH_QUERY_NEXT_KEY] = request.GET[DBMI_AUTH_QUERY_NEXT_KEY]

    # Add client ID
    state[DBMI_AUTH_QUERY_CLIENT_ID_KEY] = client_id

    # Create a token for state
    token = secrets.token_urlsafe(32)
    state["state"] = token

    # Add additional state if necessary
    provider.set_state(request, state)

    # Build authorize URL with base64 encoded state querystring
    authorize_url = provider.get_authorize_url(
        request,
        base64.urlsafe_b64encode(urlencode(state).encode('utf-8')).decode('utf-8')
    )

    # Create the response
    response = redirect(authorize_url)

    # Encrypt the state querystring for storing in cookie
    keys = [settings.SECRET_KEY] + dbmi_settings.AUTH_ENCRYPTION_KEYS
    fernet = MultiFernet([Fernet(derive_fernet_key(k)) for k in keys])
    state_enc = fernet.encrypt(urlencode(state).encode('utf-8')).decode('utf-8')

    # Place a cookie with state
    response.set_cookie(
        DBMI_AUTH_STATE_COOKIE_NAME,
        state_enc,
        max_age=2592000,
        domain=dbmi_settings.JWT_COOKIE_DOMAIN,
        secure=True,
        httponly=True,
        samesite="Lax"
    )

    # Redirect to auth provider
    return response


def check_state(request):
    """
    This endpoint is called upon callback from the auth provider and will
    retrieve the local state as well as the state in the query of the
    returning call to verify that they match. If they do not match, or an
    error is encountered while fetching state, the check is failed. Returns
    a tuple of check status, the state object.

    :param request: The current HTTP request object
    :type request: HttpRequest
    :returns: Whether the state matches or not, the state object
    :rtype: bool, dict
    """
    # Fetch some of the request parameters
    state = None
    try:
        # Check for state in request (not used for refreshes or impersonations)
        if not request.GET.get("state"):

            # Nothing to compare, let it pass
            logger.warning(f"No 'state' found in request, allowing login")
            return True, None

        # Get query from state in returning call
        return_state = QueryDict(base64.urlsafe_b64decode(request.GET["state"].encode('utf-8')).decode('utf-8'))

        # Get encrypted state in cookies
        state_enc = request.COOKIES.get(DBMI_AUTH_STATE_COOKIE_NAME)

        # Decrypt the state cookie
        keys = [settings.SECRET_KEY] + dbmi_settings.AUTH_ENCRYPTION_KEYS
        fernet = MultiFernet([Fernet(derive_fernet_key(k)) for k in keys])
        state = QueryDict(fernet.decrypt(state_enc.encode('utf-8')).decode('utf-8'), mutable=True)

        # Compare state tokens
        token = next(iter(state.pop("state")))
        if token != return_state["state"]:
            logger.error('Auth error: mismatched state',
                extra={
                    'request': request, 'state': state, 'token': token,
                    'auth_error': request.GET.get('error'),
                    'auth_error_description': request.GET.get('error_description')
                }
            )

            # Fail the check
            return False, state

        return True, state

    except InvalidToken as e:
        logger.error(f'Failed to decrypt cookie state: {e}', exc_info=True)

    except Exception as e:
        logger.error(
            f'Failed to load query(s): {e}',
            exc_info=True,
            extra={'request': request}
        )

    return False, state


def callback(request):
    """
    This endpoint is called by the auth provider with a code that lets us
    know the user logged into their Identity Provider successfully.
    We need to use the code to gather the user information from the auth
    provider and establish the DBMI_JWT cookie containing their valid JWT.

    :param request: The current HTTP request object
    :type request: HttpRequest
    """
    logger.debug("Callback")

    # Retrieve and check state
    matched, state = check_state(request)

    # Handle failure
    if not matched:
        return login_failure(
            request,
            "The authentication provider returned mismatching state. Please retry login.",
            state,
        )

    # This is a code passed back from provider that is used to retrieve a token (Which is used to retrieve user info).
    code = request.GET.get('code')
    if not code:
        logger.error('Auth code error: {} - {}'.format(
                request.GET.get('error'),
                request.GET.get('error_description')
            ),
            extra={
                'request': request, 'state': state,
                'auth_error': request.GET.get('error'),
                'auth_error_description': request.GET.get('error_description')
            }
        )
        return login_failure(
            request,
            request.GET.get('error_description'),
            state,
        )

    # Build a URL with the root URI
    callback_url = furl(request.build_absolute_uri(reverse("dbmi_login:callback")))

    # Get the client ID
    client_id = state.get(DBMI_AUTH_QUERY_CLIENT_ID_KEY) if state else next(iter(dbmi_settings.AUTH_CLIENTS.keys()))

    # Intialize the authentication backend
    provider = ProviderFactory.create(client_id, callback_url.url)
    logger.debug(f"Provider '{provider.identifier}' matched to client ID '{client_id}'")

    # Get the tokens
    id_token, access_token = provider.get_tokens(request, code)
    if not id_token or not access_token:
        logger.error(
            "No id or access tokens returned for user, cannot proceed",
        )
        return login_failure(
            request,
            "The user's tokens could not be fetched from the authentication provider.",
            state,
        )

    # Get email
    email = provider.get_user_email(request, access_token)
    if not email:
        logger.error(
            "No email returned for user info, cannot proceed"
        )
        return login_failure(
            request,
            "The user's email address could not be fetched from the authentication provider.",
            state,
        )

    # Redirect the user to the page they originally requested.
    if hasattr(dbmi_settings, "LOGIN_REDIRECT_KEY") and state.get(dbmi_settings.LOGIN_REDIRECT_KEY):
        redirect_url = state.get(dbmi_settings.LOGIN_REDIRECT_KEY)

    elif hasattr(dbmi_settings, "LOGIN_REDIRECT_URL"):
        redirect_url = dbmi_settings.LOGIN_REDIRECT_URL

    else:
        redirect_url = request.build_absolute_uri(reverse("dbmi_login:jwt"))

    logger.debug("Redirecting user to: {}".format(redirect_url))

    # Build response
    response = redirect(redirect_url)

    # Set the JWT into a cookie in the response.
    response.set_cookie(
        dbmi_settings.JWT_COOKIE_NAME,
        id_token,
        domain=dbmi_settings.JWT_COOKIE_DOMAIN,
        secure=True,
        httponly=True,
        samesite="Lax"
    )

    # Delete state cookie
    response.delete_cookie(DBMI_AUTH_STATE_COOKIE_NAME, domain=dbmi_settings.JWT_COOKIE_DOMAIN)

    return response


def login_failure(request, error, query=None):
    """
    This method builds and returns a response intended to inform the user
    of a login failure and provide remediation via retry.

    :param request: The current HTTP request object
    :type request: HttpRequest
    :param error: The error message to display
    :type error: str
    :param query: The original query as a QueryDict to include with subsequent login attempts
    :type query: QueryDict, defaults to None
    """
    # Get querystring
    querystring = f"?{query.urlencode('/')}" if query else ""

    # Get the backup authentication URL
    auth_url = furl(
        f"{request.build_absolute_uri(reverse('dbmi_login:authorize'))}{querystring}"
    )
    logger.debug(f"Backup auth URL: {auth_url.url}")

    # Set context for error page
    context = {
        "error_description": error,
        "retry_url": auth_url.url
    }

    # Render the error page
    return render(request, 'dbmi_client/login/error.html', context)

def logout(request):
    """
    This endpoint terminates a user's logged in session and redirects them
    accordingly. The session is both terminated via auth provider as well
    as locally by deleting the cookie containing their JWT.

    :param request: The current HTTP request object
    :type request: HttpRequest
    """
    # See if they are logged in
    if validate_request(request):

        # Redirect the user to the logout page
        next_url = furl(request.build_absolute_uri(reverse("dbmi_login:logout")))

        # Get the client ID
        client_id = get_jwt_client_id(request)

        # Intialize the authentication backend
        provider = ProviderFactory.create(client_id, None)
        logger.debug(f"Provider '{provider.identifier}' matched to client ID '{client_id}'")

        # Get the logout URL
        url = provider.get_logout_url(request, next_url.url)

        # Log the URL
        logger.debug(f"Logout URL: {url}")

        # Ensure the request is cleared of user state
        django_auth.logout(request)

        # Create the response
        response = redirect(url)

        # Look for next url
        if request.GET.get(dbmi_settings.LOGOUT_REDIRECT_KEY):

            # Get the passed URL
            logger.debug('Will log user out and redirect to: {}'.format(
                request.GET.get(dbmi_settings.LOGOUT_REDIRECT_KEY)
            ))

            # Set the next URL into a cookie in the response.
            response.set_cookie(
                DBMI_AUTH_LOGOUT_COOKIE_NAME,
                request.GET.get(dbmi_settings.LOGOUT_REDIRECT_KEY),
                domain=dbmi_settings.JWT_COOKIE_DOMAIN,
                secure=True,
                httponly=True,
                samesite="Lax"
            )

        # Set the URL and purge cookies
        response.delete_cookie(dbmi_settings.JWT_COOKIE_NAME, domain=dbmi_settings.JWT_COOKIE_DOMAIN)

        return response

    else:
        logger.debug("User has been logged out, sending to logout page")

        # Check cookies for the redirect URL
        next_url = request.COOKIES.get(DBMI_AUTH_LOGOUT_COOKIE_NAME)

        # Look for next url
        if next_url:

            # Prepare the response
            response = redirect(next_url)

            # Delete cookie with redirect URL
            response.delete_cookie(DBMI_AUTH_LOGOUT_COOKIE_NAME, domain=dbmi_settings.JWT_COOKIE_DOMAIN)

            # Send them off
            logger.debug('Will redirect logged out user to: {}'.format(next_url))
            return response

        # Set context
        context = {
            "title": dbmi_settings.AUTHN_TITLE,
        }

        # Render the logout landing page
        return render(request, "dbmi_client/login/logout.html", context)



import inspect
import pkgutil
from pathlib import Path
from importlib import import_module

from django.conf import settings
from dbmi_client.settings import dbmi_settings

import logging
logger = logging.getLogger(__name__)


class ProviderFactory(object):
    """
    This class manages the creation of Provider instances based on the identifier/provider
    and client ID.
    """
    @classmethod
    def provider_for_client_id(cls, client_id):
        """
        Returns the provider identifier for the given client ID.

        :param client_id: The client ID of the auth provider to check
        :type client_id: str
        :return: The identifier of the provider
        :rtype: str
        """
        try:
            # Match the client ID to an authentication provider
            return dbmi_settings.AUTH_CLIENTS[client_id]["PROVIDER"]

        except Exception:
            raise Exception(f"No matching authentication provider for client id: '{client_id}'")

    @classmethod
    def create(cls, client_id, callback_url, **kwargs):

        identifier = None
        try:
            # Match the client ID to an authentication provider
            configuration = dbmi_settings.AUTH_CLIENTS[client_id]
            identifier = configuration["PROVIDER"]

            # Get the class for the provider
            provider_class = next(
                provider for provider in ProviderFactory.providers()
                if getattr(provider, 'identifier') == identifier
            )

            # Instantiate it
            provider = provider_class(**{
                'domain': configuration["DOMAIN"],
                'client_id': client_id,
                'client_secret': configuration["CLIENT_SECRET"],
                'scope': configuration["SCOPE"],
                'callback_url': callback_url,
                **kwargs
            })

            return provider

        except StopIteration:
            raise Exception(f"No implemented authentication provider for: '{identifier}' / '{client_id}'")

        except Exception:
            raise Exception(f"No matching authentication provider for client id: '{client_id}'")

    #
    # META
    #

    @classmethod
    def providers(cls, filter=None):
        """
        This method returns a complete list of all subclasses inherited from Provider. The list contains an
        instance of each class. If specified, the filter will limit the set of subclasses returned based
        on a key-value check on the class properties.
        :param filter: A dictionary defining a filter to apply to the returned list of subclasses.
        :return: list
        """
        provider_class = getattr(import_module('dbmi_client.provider.provider'), "Provider")
        subclasses = []
        for (_, name, _) in pkgutil.iter_modules([str(Path(__file__).parent)]):
            imported_module = import_module('.' + name, package='dbmi_client.provider')
            for i in dir(imported_module):
                attribute = getattr(imported_module, i)

                # Ensure this is a valid subclass of provider
                if inspect.isclass(attribute) and issubclass(attribute, provider_class) and attribute not in subclasses:

                    # Check for child providers
                    if hasattr(attribute, 'providers'):
                        for provider_subclass in getattr(attribute, 'providers')(filter=filter):
                            if inspect.isclass(provider_subclass) and issubclass(provider_subclass, provider_class) \
                                    and provider_subclass not in subclasses:
                                subclasses.append(provider_subclass)
                            else:
                                logger.warning('Provider "{}" returned a child provider "{}" that did not qualify '
                                               'as provider'.format(attribute, provider_subclass))

                    # Ensure a base set of required properties exist
                    if not getattr(attribute, 'identifier', False):
                        continue

                    # Check filter
                    if filter:
                        matched = False
                        for key, value in filter.items():
                            if getattr(attribute, key, None) != value:
                                break
                        else:
                            matched = True

                        # Check result
                        if not matched:
                            continue

                    # Passed all tests, this subclass is a valid Provider
                    subclasses.append(attribute)

        return subclasses



import sys
import logging


def config(service, sentry=True, root_level=logging.WARNING):

    # Set the standard configurations
    config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "console": {
                "format": f"[DJANGO] - [{service.upper()}] - [%(asctime)s][%(levelname)s]"
                f"[%(name)s.%(funcName)s:%(lineno)d] - %(message)s",
            },
        },
        "handlers": {
            "console": {
                "level": "DEBUG",
                "class": "logging.StreamHandler",
                "formatter": "console",
                "stream": sys.stdout,
            },
        },
        "root": {
            "handlers": ["console"],
            "level": logging.getLevelName(root_level),
        },
        "loggers": {
            "django": {
                "handlers": ["console"],
                "level": "WARNING",
                "propagate": True,
            },
            "django.request": {
                "handlers": ["console"],
                "level": "ERROR",
                "propagate": True,
            },
            'py.warnings': {
                'handlers': ['console'],
                'level': 'WARNING',
                'propagate': True
            },
        },
    }

    # Add Sentry/Raven
    if sentry:

        # Add the handler
        config["handlers"]["sentry"] = {
            "level": "ERROR",  # To capture more than ERROR, change to WARNING, INFO, etc.
            "class": "raven.contrib.django.raven_compat.handlers.SentryHandler",
            "tags": {"custom-tag": "x"},
        }
        config["root"]["handlers"].append("sentry")

        # Add the loggers
        config["loggers"]["raven"] = {
            "level": "WARNING",
            "handlers": ["console"],
            "propagate": False,
        }

        config["loggers"]["sentry.errors"] = {
            "level": "WARNING",
            "handlers": ["console"],
            "propagate": False,
        }

    return config


def add_level(levelName, levelNum, methodName=None):
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example
    -------
    >>> addLoggingLevel('TRACE', logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("TRACE")
    >>> logging.getLogger(__name__).trace('that worked')
    >>> logging.trace('so did this')
    >>> logging.TRACE
    5

    """
    if not methodName:
        methodName = levelName.lower()

    if hasattr(logging, levelName):
        raise AttributeError("{} already defined in logging module".format(levelName))
    if hasattr(logging, methodName):
        raise AttributeError("{} already defined in logging module".format(methodName))
    if hasattr(logging.getLoggerClass(), methodName):
        raise AttributeError("{} already defined in logger class".format(methodName))

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def logForLevel(self, message, *args, **kwargs):
        if self.isEnabledFor(levelNum):
            self._log(levelNum, message, args, **kwargs)

    def logToRoot(message, *args, **kwargs):
        logging.log(levelNum, message, *args, **kwargs)

    logging.addLevelName(levelNum, levelName)
    setattr(logging, levelName, levelNum)
    setattr(logging.getLoggerClass(), methodName, logForLevel)
    setattr(logging, methodName, logToRoot)



from __future__ import absolute_import, unicode_literals

import os
import warnings
import logging

from django.conf import settings
from django.test.signals import setting_changed

from dbmi_client import environment as env

# Always import this module as follows:
# from dbmi_client import settings [as dbmi_settings]

# Don't import directly CONFIG or PANELs, or you will miss changes performed
# with override_settings in tests.

# Set service URLs based on environment
DBMI_ENVIRONMENTS = {
    "prod": {
        "AUTHN_URL": "https://authentication.dbmi.hms.harvard.edu",
        "AUTHZ_URL": "https://authorization.dbmi.hms.harvard.edu",
        "REG_URL": "https://registration.dbmi.hms.harvard.edu",
        "FILESERVICE_URL": "https://files.dbmi.hms.harvard.edu",
        "JWT_AUTHZ_NAMESPACE": "https://authorization.dbmi.hms.harvard.edu",
    },
    "dev": {
        "AUTHN_URL": "https://authentication.aws.dbmi-dev.hms.harvard.edu",
        "AUTHZ_URL": "https://authorization.aws.dbmi-dev.hms.harvard.edu",
        "REG_URL": "https://registration.aws.dbmi-dev.hms.harvard.edu",
        "FILESERVICE_URL": "https://fileservice.aws.dbmi-dev.hms.harvard.edu",
        "JWT_AUTHZ_NAMESPACE": "https://authorization.aws.dbmi-dev.hms.harvard.edu",
    },
}

CONFIG_DEFAULTS = {
    # The identifier for this service and/or project
    "CLIENT": None,
    # Client options, assume production environment
    "ENVIRONMENT": "prod",
    # Set prod URLs
    "AUTHN_URL": None,
    "AUTHZ_URL": None,
    "REG_URL": None,
    "FILESERVICE_URL": None,
    # Optionally disable logging
    "LOGGER_NAME": "dbmi_client",
    "ENABLE_LOGGING": True,
    "LOG_LEVEL": logging.WARNING,
    # Universal login screen branding
    "AUTHN_TITLE": "DBMI Client",
    "AUTHN_ICON_URL": "https://authentication.dbmi.hms.harvard.edu/static/hms_logo.svg",
    "AUTHN_COLOR": "#bc262b",
    "AUTHN_BACKGROUND": None,
    # AuthZ groups/roles/permissions
    "AUTHZ_ADMIN_GROUP": "dbmi-admin",
    "AUTHZ_ADMIN_PERMISSION": "MANAGE",
    "AUTHZ_USER_GROUP": "dbmi-user",
    "AUTHZ_REPORTER_CLASS": "dbmi_client.authz.AuthorizationReporter",
    # JWT bits
    "JWT_AUTHZ_NAMESPACE": None,
    "JWT_HTTP_PREFIX": "JWT ",
    "JWT_COOKIE_NAME": "DBMI_JWT",
    "JWT_COOKIE_DOMAIN": ".dbmi.hms.harvard.edu",
    # Authentication provider details
    "AUTH_CLIENTS": None,
    "AUTH_ENCRYPTION_KEYS": [],
    # Configurations surrounding usage of a local Django user model
    "USER_MODEL_ENABLED": False,
    "USER_MODEL_AUTOCREATE": True,
    # These configurations are specific to DRF related auth/permissions
    "DRF_OBJECT_OWNER_KEY": "user",
    # Jira support desk properties
    "JIRA_ORGANIZATION": "hms-dbmi",
    "JIRA_SERVICE_DESK": None,
    "JIRA_SERVICE_DESK_EMAIL": None,
    "JIRA_USERNAME": None,
    "JIRA_TOKEN": None,
    # Filservice
    "FILESERVICE_TOKEN": None,  # The authentication token to be used with Fileservice
    "FILESERVICE_BUCKETS": [],  # The name of the S3 buckets Fileservice should use
    "FILESERVICE_GROUP": None,  # Typically would be the same as CLIENT
    "FILESERVICE_ADMINS": [],  # A list of Fileservice users and/or user emails that should administer group
    # Login settings
    "LOGIN_REDIRECT_KEY": "next",  # The query parameter key specifying where logged in users should be sent
    "LOGIN_REDIRECT_URL": None,  # The post-login URL to send users to if not specified by 'LOGIN_REDIRECT_KEY'
    # Logout settings
    "LOGOUT_REDIRECT_KEY": "next",  # The query parameter key specifying where logged out users should be sent
    "LOGOUT_REDIRECT_URL": None,  # The post-logout URL to send users to if not specified by 'LOGOUT_REDIRECT_KEY'
}

# List of settings that cannot be defaulted and must be user-defined
REQUIRED_SETTINGS = ("CLIENT", "AUTH_CLIENTS", )

# List of settings that have been removed
REMOVED_SETTINGS = (
    "AUTH0_TENANT", "AUTH0_CLIENT_ID", "AUTH0_SECRET", "AUTH0_CLIENTS", "AUTH0_TENANTS", "AUTH0_SCOPE", "AUTH0_DOMAIN"
)


class DBMISettings(object):
    """
    A settings object, that allows DBMI Client settings to be accessed as properties.
    For example:
        from dbmi_client.settings import dbmi_settings
        print(dbmi_settings.AUTHZ_ADMIN_GROUP)
    Any setting with string import paths will be automatically resolved
    and return the class, rather than the string literal.
    """

    def __init__(self, defaults=None):
        self.defaults = defaults or CONFIG_DEFAULTS
        self._cached_attrs = set()

    @property
    def user_settings(self):

        # Check to see if user configs have been loaded or not
        if not hasattr(self, "_user_settings"):

            try:
                # Load user-specified configurations
                user_settings = getattr(settings, "DBMI_CLIENT_CONFIG", {})

                # Update the client config with pre-defined environment URLs, etc
                if user_settings.get("ENVIRONMENT") in DBMI_ENVIRONMENTS:
                    env_settings = DBMI_ENVIRONMENTS[user_settings.get("ENVIRONMENT")]
                    user_settings.update({k: v for k, v in env_settings.items() if k not in user_settings})

                else:
                    # Check for them in environment
                    for key in DBMI_ENVIRONMENTS["prod"].keys():
                        if env.get_str("DBMI_{}".format(key.upper())):
                            user_settings[key] = env.get_str("DBMI_{}".format(key.upper()))

                # Check them
                self._user_settings = self.__check_user_settings(user_settings)

            except Exception as e:
                raise SystemError("DBMI Client settings are invalid: {}".format(e))

        return self._user_settings

    def __getattr__(self, attr):

        # Any attribute must be in either required settings or defaults
        if attr not in REQUIRED_SETTINGS and attr not in self.defaults:
            raise AttributeError("Invalid DBMI setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:

            # Fall back to defaults
            val = self.defaults[attr]

        # Cache the result
        self._cached_attrs.add(attr)
        setattr(self, attr, val)

        return val

    def __check_user_settings(self, user_settings):
        SETTINGS_DOC = "https://github.com/hms-dbmi/django-dbmi-client"

        for setting in REMOVED_SETTINGS:
            if setting in user_settings:
                raise RuntimeError(
                    "The '%s' setting has been removed. Please refer to '%s' for available settings."
                    % (setting, SETTINGS_DOC)
                )

        if "CLIENT" not in user_settings:
            raise AttributeError("CLIENT configuration must be set")

        # Check auth configuration(s)
        if type(user_settings["AUTH_CLIENTS"]) is not dict:
            raise AttributeError("AUTH_CLIENTS configuration must be set as a dictionary")

        # Check each auth configuration
        for _, configuration in user_settings["AUTH_CLIENTS"].items():

            # Each client needs a minimum of a JWKS URL
            if not configuration.get("JWKS_URL") and not configuration.get("CLIENT_SECRET"):
                raise AttributeError("Each auth client required a configured JWKS_URL or CLIENT_SECRET")

            # If the login app is enabled, more is required for each client
            if "dbmi_client.login" in settings.INSTALLED_APPS:

                # Check secret
                if not configuration.get("PROVIDER"):
                    raise AttributeError("Each auth client required a configured PROVIDER: 'auth0' or 'cognito'")

                # Check secret
                if not configuration.get("CLIENT_SECRET"):
                    raise AttributeError("Each auth client required a configured CLIENT_SECRET")

                # Check secret
                if not configuration.get("DOMAIN"):
                    raise AttributeError("Each auth client required a configured DOMAIN")

                # Check secret
                if not configuration.get("SCOPE"):
                    raise AttributeError("Each auth client required a configured SCOPE")

        # Ensure environment is set and if not prod or dev, ensure service URLs are provided
        if (
            "ENVIRONMENT" in user_settings
            and user_settings["ENVIRONMENT"].lower() != "prod"
            and user_settings["ENVIRONMENT"].lower() != "dev"
        ):
            warnings.warn("ENVIRONMENT is not set to production, this should be a test environment", ResourceWarning)

            missing_urls = []
            if "AUTHN_URL" not in user_settings:
                # Check environment
                if os.environ.get("DBMI_AUTHN_URL"):
                    user_settings["AUTHN_URL"] = os.environ.get("DBMI_AUTHN_URL")
                else:
                    missing_urls.append("AUTHN_URL")

            if "AUTHZ_URL" not in user_settings:
                # Check environment
                if os.environ.get("DBMI_AUTHZ_URL"):
                    user_settings["AUTHZ_URL"] = os.environ.get("DBMI_AUTHZ_URL")
                else:
                    missing_urls.append("AUTHZ_URL")

            if "REG_URL" not in user_settings:
                # Check environment
                if os.environ.get("DBMI_REG_URL"):
                    user_settings["REG_URL"] = os.environ.get("DBMI_REG_URL")
                else:
                    missing_urls.append("REG_URL")
            if missing_urls:
                raise AttributeError("{} configuration(s) must be set".format(missing_urls))

            if "FILESERVICE_URL" not in user_settings:
                # Check environment
                if os.environ.get("DBMI_FILESERVICE_URL"):
                    user_settings["FILESERVICE_URL"] = os.environ.get("DBMI_FILESERVICE_URL")
                else:
                    warnings.warn("FILESERVICE_URL is not set, Fileservice requests will fail", ResourceWarning)

            if (
                "FILESERVICE_GROUP" in user_settings
                or "FILESERVICE_BUCKETS" in user_settings
                or "FILESERVICE_ADMINS" in user_settings
                or "FILESERVICE_TOKEN" in user_settings
            ) and (
                "FILESERVICE_GROUP" not in user_settings
                or "FILESERVICE_BUCKETS" not in user_settings
                or "FILESERVICE_ADMINS" not in user_settings
                or "FILESERVICE_TOKEN" not in user_settings
            ):
                warnings.warn(
                    "FILESERVICE_GROUP, FILESERVICE_BUCKETS, FILESERVICE_ADMINS and FILESERVICE_TOKEN must be "
                    "defined for Fileservice usage",
                    ResourceWarning,
                )

            if "JWT_AUTHZ_NAMESPACE" not in user_settings:
                warnings.warn(
                    "JWT_AUTHZ_NAMESPACE is not set, JWT claims will not be inspected " "for groups/roles/permissions",
                    ResourceWarning,
                )

        if "JWT_HTTP_PREFIX" in user_settings:
            warnings.warn("Changing JWT_HTTP_PREFIX breaks compatibility with DBMI services", ResourceWarning)

        if "JWT_COOKIE_NAME" in user_settings:
            warnings.warn("Changing JWT_COOKIE_NAME breaks compatibility with DBMI services", ResourceWarning)

        return user_settings

    def reload(self):
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        if hasattr(self, "_user_settings"):
            delattr(self, "_user_settings")

    def __dir__(self):

        # Return defaults and user configs patched over
        attrs = dict()
        attrs.update(self.defaults)
        attrs.update(self.user_settings)

        return attrs.keys()

    @property
    def USER_MODEL_ENABLED(self):

        try:
            # Check if present in user settings
            return self.user_settings["USER_MODEL_ENABLED"]

        except KeyError:

            # Else, return true if the client has added either backend classes to AUTHENTICATION_BACKENDS
            return (
                "dbmi_client.authn.DBMIModelAuthenticationBackend" in settings.AUTHENTICATION_BACKENDS
                or "dbmi_client.authn.DBMIUsersModelAuthenticationBackend" in settings.AUTHENTICATION_BACKENDS
                or "dbmi_client.authn.DBMIAdminModelAuthenticationBackend" in settings.AUTHENTICATION_BACKENDS
            )

    def get_logger(self):
        """
        Returns the logger and manages whether logs propogate or not depending on user configs
        :return: logger
        """
        logger = logging.getLogger(__name__)

        # Check if disabled
        if not self.ENABLE_LOGGING:
            logger.propagate = False

        return logger


# Create the instance by which the settings should be accessed
dbmi_settings = DBMISettings(CONFIG_DEFAULTS)

# Configure logging
logger = logging.getLogger(dbmi_settings.LOGGER_NAME)
logger.disabled = not dbmi_settings.ENABLE_LOGGING
logger.setLevel(dbmi_settings.LOG_LEVEL)

# Set the sentry client if Raven is used
if "raven.contrib.django.raven_compat" in settings.INSTALLED_APPS:
    setattr(settings, "SENTRY_CLIENT", "dbmi_client.sentry.DBMISentryClient")


def reload_dbmi_settings(*args, **kwargs):
    setting = kwargs["setting"]
    if setting == "DBMI_CLIENT_CONFIG":
        dbmi_settings.reload()


setting_changed.connect(reload_dbmi_settings)



class Provider(object):
    """
    The provider class encapsulates authentication provider behaviors
    and routines.
    """
    identifier = None
    domain = None
    client_id = None
    client_secret = None
    scope = None
    callback_url = None

    def __init__(self, domain, client_id, client_secret, scope, callback_url):
        self.domain = domain
        self.client_id = client_id
        self.client_secret = client_secret
        self.scope = scope
        self.callback_url = callback_url

    def set_state(self, request, state):
        """
        This method allows the provider instance to add to or modify the state
        object that is passed along with the '/authorize' request. This allows
        the service to pass parameters through login to the redirect to the
        calling service.

        :param request: The current request
        :type request: HttpRequest
        :param state: The current state object
        :type state: dict
        """
        pass

    def get_authorize_url(self, request, state):
        """
        This method returns the URL to be used for the authorize endpoint. This typically redirects users
        to the authentication provider's hosted login.

        :param request: The current request object
        :type request: HttpRequest
        :param state: An object containing state as well as anything else that needs to pass-through
        :type state: dict
        :returns: The URL to redirect users to for initial login
        :rtype: furl
        """
        raise NotImplementedError(f"This method should be implemented by subclasses")

    def get_tokens(self, request, code):
        """
        This method takes an authorization code and fetches the ID and access tokens from the authentication provider.

        :param request: The current request object
        :type request: HttpRequest
        :param code: The code returned upon user's login
        :type code: str
        :returns: A tuple containing the ID token and the access token
        :rtype: str, str
        """
        raise NotImplementedError(f"This method should be implemented by subclasses")

    def get_user_email(self, request, access_token):
        """
        This method calls the authentication provider to return the currently logged in user's email address

        :param request: The current request object
        :type request: HttpRequest
        :param access_token: The user's access token
        :type access_token: str
        :returns: The user's email address
        :rtype: str
        """
        raise NotImplementedError(f"This method should be implemented by subclasses")

    def get_logout_url(self, request, next_url):
        """
        This method builds and returns the URL to redirect users to when processing a logout from the authentication
        provider. The next_url parameter allows specifying where the user will be redirected upon logout.

        :param request: The current request object
        :type request: HttpRequest
        :param next_url: The URL to redirect recently logged out users to
        :type next_url: str
        :returns: The logout URL to send users to
        :rtype: str
        """
        raise NotImplementedError(f"This method should be implemented by subclasses")

    def is_member_of_group(self, request, group):
        """
        This method inspects the claims of the current request's JWT and returns
        whether the authentication provider has indicated membership in the
        passed group or not.

        :param request: The current request object
        :type request: HttpRequest
        :param group: The name of the group to check membership of
        :type group: str
        :returns: Whether the user belongs to the group or not
        :rtype: bool
        """
        raise NotImplementedError(f"This method should be implemented by subclasses")



import base64
import json
from furl import furl
import requests

from dbmi_client.authn import get_jwt_payload
from dbmi_client.provider.provider import Provider
from dbmi_client.login.views import DBMI_AUTH_QUERY_BRANDING_KEY
from dbmi_client.settings import dbmi_settings

import logging
logger = logging.getLogger(__name__)


class Auth0(Provider):
    """
    The provider class encapsulates authentication provider behaviors
    and routines.
    """
    identifier = "auth0"

    def set_state(self, request, state):
        """
        This method allows the provider instance to add to or modify the state
        object that is passed along with the '/authorize' request. This allows
        the service to pass parameters through login to the redirect to the
        calling service.

        :param request: The current request
        :type request: HttpRequest
        :param state: The current state object
        :type state: dict
        """
        # Check for authentication query elements
        if request.GET.get(DBMI_AUTH_QUERY_BRANDING_KEY):

            # Add it
            state[DBMI_AUTH_QUERY_BRANDING_KEY] = request.GET[DBMI_AUTH_QUERY_BRANDING_KEY]
            logger.debug(f"Passing along branding")

        else:

            # Add default UI customizations for hosted login
            branding = {}
            if dbmi_settings.AUTHN_TITLE:
                branding["title"] = dbmi_settings.AUTHN_TITLE
            if dbmi_settings.AUTHN_ICON_URL:
                branding["icon_url"] = dbmi_settings.AUTHN_ICON_URL
            if dbmi_settings.AUTHN_COLOR:
                branding["color"] = dbmi_settings.AUTHN_COLOR
            if dbmi_settings.AUTHN_BACKGROUND:
                branding["background"] = dbmi_settings.AUTHN_BACKGROUND

            # Add it
            state[DBMI_AUTH_QUERY_BRANDING_KEY] = base64.urlsafe_b64encode(json.dumps(branding).encode('utf-8')).decode('utf-8')
            logger.debug(f"Using default branding")


    def get_authorize_url(self, request, state):
        """
        This method returns the URL to be used for the authorize endpoint. This typically redirects users
        to the authentication provider's hosted login.

        :param request: The current request object
        :type request: HttpRequest
        :param state: The string to use as the state parameter
        :type state: str
        :returns: The URL to redirect users to for initial login
        :rtype: str
        """
        url = furl(
            f"https://{self.domain}/authorize"
        )

        # Add required parameters
        url.query.params.add("response_type", "code")
        url.query.params.add("client_id", self.client_id)
        url.query.params.add("redirect_uri", self.callback_url)
        url.query.params.add("scope", self.scope)
        url.query.params.add("state", state)

        return url.url

    def get_tokens(self, request, code):
        """
        This method takes an authorization code and fetches the ID and access tokens from the authentication provider.

        :param request: The current request object
        :type request: HttpRequest
        :param code: The code returned upon user's login
        :type code: str
        :returns: A tuple containing the ID token and the access token
        :rtype: str, str
        """
        # This is the URL we post the code to in order to get token.
        url = furl(f'https://{self.domain}/oauth/token')

        # Set request headers
        headers = {'Content-Type': 'application/json'}

        # Information we pass to identify us and our request.
        payload = {
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'redirect_uri': self.callback_url,
            'code': code,
            'grant_type': 'authorization_code'
        }

        # Make the request
        response = requests.post(url.url, json=payload, headers=headers)

        # Check response
        if not response.ok:
            logger.debug(
                f"Token exchange request failure: {url} / {response.status_code} / {response.content}"
            )
            logger.error(
                'Auth failure: Failed to exchange token',
                extra={
                    'url': url.url,
                    'response': response.content,
                    'status': response.status_code,
                }
            )

            # Cannot proceed without token
            return None, None

        # Return tokens
        return response.json()["id_token"], response.json()["access_token"]

    def get_user_email(self, request, access_token):
        """
        This method calls the authentication provider to return the currently logged in user's email address

        :param request: The current request object
        :type request: HttpRequest
        :param access_token: The user's access token
        :type access_token: str
        :returns: The user's email address
        :rtype: str
        """
        # Set the user profile endpoint URL
        url = furl(f'https://{self.domain}/userinfo')

        # Set query parameters
        params = {"access_token": access_token}

        # Make the request
        response = requests.get(url.url, params=params)

        # Check response
        if not response.ok:
            logger.debug(
                f"Profile request failure: {url} / {response.status_code} / {response.content}"
            )
            logger.error(
                'Auth failure: Failed to retrieve user profile',
                extra={
                    'url': url.url,
                    'response': response.content,
                    'status': response.status_code,
                }
            )
            return None

        # Parse email from response
        return response.json()["email"]

    def get_logout_url(self, request, next_url):
        """
        This method builds and returns the URL to redirect users to when processing a logout from the authentication
        provider. The next_url parameter allows specifying where the user will be redirected upon logout.

        :param request: The current request object
        :type request: HttpRequest
        :param next_url: The URL to redirect recently logged out users to
        :type next_url: str
        :returns: The logout URL to send users to
        :rtype: str
        """
        # Set the logout URL
        url = furl(f'https://{self.domain}/v2/logout')

        # Add the client ID and redirect
        url.query.params.add('client_id', self.client_id)
        url.query.params.set('returnTo', next_url)

        return url.url

    def is_member_of_group(self, request, group):
        """
        This method inspects the claims of the current request's JWT and returns
        whether the authentication provider has indicated membership in the
        passed group or not.

        :param request: The current request object
        :type request: HttpRequest
        :param group: The name of the group to check membership of
        :type group: str
        :returns: Whether the user belongs to the group or not
        :rtype: bool
        """
        # Get payload
        payload = get_jwt_payload(request)

        # Check claims for the groups list
        groups = payload.get(dbmi_settings.JWT_AUTHZ_NAMESPACE, {}).get("groups", [])

        # Iterate groups
        for _group in groups:

            # Check type
            if type(_group) is str and _group == group:
                return True

            if type(_group) is dict and _group.get("name") == group:
                return True

        return False



import os
import re
from io import open

from setuptools import find_packages, setup


def read(f):
    return open(f, "r", encoding="utf-8").read()


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(
        os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            os.path.join(package, "__init__.py"),
        )
    ).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


version = get_version("dbmi_client")

setup(
    name="django-dbmi-client",
    version=version,
    url="https://github.com/hms-dbmi/django-dbmi-client",
    author="HMS DBMI Tech-core",
    author_email="dbmi-tech-core@hms.harvard.edu",
    packages=find_packages(exclude=["tests*"]),
    license="Creative Commons Attribution-Noncommercial-Share Alike license",
    install_requires=read("requirements.txt").splitlines(),
    tests_require=read("requirements-test.txt").splitlines(),
    extras_require={
        "test": read("requirements-test.txt").splitlines(),
        "dev": read("requirements-dev.txt").splitlines(),
    },
    test_suite="dbmi_client.tests.runtests.main",
    include_package_data=True,
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3",
        "Framework :: Django :: 4",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",  # example license
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
)



import mock
import re
import responses

from django.test import TestCase
from django.http import HttpResponse
from django.contrib.auth.models import User, AnonymousUser
from django.test.client import RequestFactory
from dbmi_client import authz
from dbmi_client.settings import dbmi_settings
from dbmi_client import auth


class TestAuth(TestCase):

    # User JWT
    fake_jwt = "somefakejwtalsdijlaiwjdlijawlidjasdhgashgd"

    @classmethod
    def setUpClass(cls):

        # Patch JWT checker
        cls.jwt_patcher = mock.patch("dbmi_client.authn.validate_request")
        cls.mock_jwt = cls.jwt_patcher.start()
        cls.mock_jwt.return_value = True

        # Set a user and request factory
        cls.rf = RequestFactory()
        cls.user = User.objects.create_user(
            username="test_auth_user", email="testauthuser@dbmiauth.local", password="top_secret"
        )

        # Set the authorization server URL
        cls.authz_url_pattern = re.compile(dbmi_settings.AUTHZ_URL + r".*")

    @classmethod
    def tearDownClass(cls):

        # Disable patcher
        cls.jwt_patcher.stop()

    def build_request(self, path, method="get", user=True):

        # Build the request
        request = getattr(self.rf, method)(path, HTTP_AUTHORIZATION=TestAuth.fake_jwt)

        # Set empty session
        setattr(request, "session", {})

        # Check for a user
        if user:
            # Add the user
            request.user = self.user
            request.COOKIES[dbmi_settings.JWT_COOKIE_NAME] = "JWT {}".format(TestAuth.fake_jwt)

        else:
            request.user = AnonymousUser()

        return request

    def setUp(self):

        # Setup each test
        pass

    @responses.activate
    @mock.patch("dbmi_client.authz.has_a_permission")
    def test_auth_mock_(self, mock_has_a_permission):

        # Set the response
        mock_has_a_permission.return_value = True

        # Build the request
        request = self.build_request("/some/page/")

        # Run it
        has_authz = authz.has_a_permission(request, self.user.email, "item", ["admin", "read"], check_parents=True)

        # Check it
        self.assertTrue(has_authz)

    @responses.activate
    def test_auth_response_(self):

        # Build the request
        request = self.build_request("/some/page/")

        # Build the response handler
        responses.add(responses.GET, self.authz_url_pattern, json={"error": "not found"}, status=404)

        # Build the call
        has_authz = authz.has_permission(request, self.user.email, "item", "admin", check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertFalse(has_authz)

    @responses.activate
    def test_auth_dbmi_user_auth(self):

        # Build the request
        request = self.build_request("/some/page/", user=True)

        # Create a fake view
        view = mock.MagicMock(return_value=HttpResponse("Some page content"))

        # Run.
        decorated = auth.dbmi_user(view)
        response = decorated(request, *[], **{})

        # Check.
        # View was called.
        view.assert_called_once_with(request, *[], **{})

        # Ensure the user is allowed
        self.assertEqual(response.status_code, 200)

    @responses.activate
    @mock.patch("dbmi_client.authn.login_redirect")
    def test_auth_dbmi_user_no_auth(self, mock_login_redirect):

        # Build the request
        request = self.build_request("/some/page/", user=False)

        # Create a fake view
        view = mock.MagicMock(return_value=HttpResponse("Some page content"))

        # Run.
        decorated = auth.dbmi_user(view)
        decorated(request, *[], **{})

        # Check.
        # View was never reached.
        view.assert_not_called()

        # Ensure the user is redirected to login.
        mock_login_redirect.assert_called_once_with(request)
